"""This module contains the core classes for the "graphic" system.

.. autosummary::
   :toctree: .

   Sprixel
   Sprite
   SpriteCollection
   Animation
"""
from pygamelib import board_items
from pygamelib import base
from pygamelib import constants
from pygamelib.assets import graphics
import time
from collections import UserDict
from uuid import uuid4
import json


class Sprixel(object):

    """
    A sprixel is the representation of 1 cell of the sprite or one cell on the Board.
    It is not really a pixel but it is the closest notion we'll have.
    A Sprixel has a background color, a foreground color and a model.
    All regular BoardItems can have use Sprixel instead of model.

    If the background color and the is_bg_transparent are None or empty strings,
    the sprixel will be automatically configured with transparent background.
    In that case, as we can really achieve transparency in the console, the sprixel will
    take the background color of whatever it is overlapping.

    :param model: The model, it can be any string. Preferrably a single character.
    :type model: str
    :param bg_color: An ANSI escape sequence to configure the background color.
    :type bg_color: str
    :param fg_color: An ANSI escape sequence to configure the foreground color.
    :type fg_color: str
    :param is_bg_transparent:
    :type is_bg_transparent:

    Example::

        player = Player(sprixel=Sprixel(
                                        '#',
                                        screen.terminal.on_color_rgb(128,56,32),
                                        screen.terminal.color_rgb(255,255,0),
                                        ))
    """

    def __init__(self, model="", bg_color="", fg_color="", is_bg_transparent=None):
        super().__init__()
        self.model = model
        self.bg_color = bg_color
        self.fg_color = fg_color
        if (
            (type(model) is not str and "FormattingString" not in str(type(model)))
            or (
                type(bg_color) is not str
                and "FormattingString" not in str(type(bg_color))
            )
            or (
                type(fg_color) is not str
                and "FormattingString" not in str(type(fg_color))
            )
        ):
            raise base.PglInvalidTypeException(
                "Sprixel(model, bg_color, fg_color): all 3 variables needs to be a "
                "string or an empty string."
            )

        if (bg_color is None or bg_color == "") and (
            is_bg_transparent is None or is_bg_transparent == ""
        ):
            self.is_bg_transparent = True
        else:
            self.is_bg_transparent = False

    def __repr__(self):
        return f"{self.bg_color}{self.fg_color}{self.model}\x1b[0m"

    def __str__(self):  # pragma: no cover
        return self.__repr__()

    def __eq__(self, other):
        if (
            self.model == other.model
            and self.bg_color == other.bg_color
            and self.fg_color == other.fg_color
        ):
            return True
        else:
            return False

    def __ne__(self, other):
        if (
            self.model != other.model
            or self.bg_color != other.bg_color
            or self.fg_color != other.fg_color
        ):
            return True
        else:
            return False

    @staticmethod
    def from_ansi(string):
        """Takes an ANSI string, parse it and return a Sprixel.

        :param string: The ANSI string to parse.
        :type string: str

        Example::

            new_sprixel = Sprixel.from_ansi(
                "\\x1b[48;2;139;22;19m\\x1b[38;2;160;26;23m▄\\x1b[0m"
            )

        .. warning:: This has mainly be tested with ANSI string generated by climage.
        """
        new_sprixel = Sprixel()
        if "[48;" in string and "[38;" in string and "▄" in string:
            (colors, end) = string.split("▄")
            new_sprixel.model = "▄"
            for e in colors.split("m"):
                if "[48;" in e:
                    new_sprixel.bg_color = f"{e}m"
                elif "[38;" in e:
                    new_sprixel.fg_color = f"{e}m"
        elif "[38;" in string and "▄" in string:
            (colors, end) = string.split("▄")
            new_sprixel.model = "▄"
            for e in colors.split("m"):
                if "[38;" in e:
                    new_sprixel.fg_color = f"{e}m"
        elif "[48;" in string and "▄" in string:
            (colors, end) = string.split("▄")
            new_sprixel.model = "▄"
            for e in colors.split("m"):
                if "[48;" in e:
                    new_sprixel.bg_color = f"{e}m"
        if new_sprixel.bg_color != "":
            new_sprixel.is_bg_transparent = False
        return new_sprixel

    @property
    def model(self):
        return self.__model

    @model.setter
    def model(self, value):
        if type(value) is str:
            self.__model = value
        else:
            raise base.PglInvalidTypeException(
                f"A Sprixel.model must be a string. {value} is not a string."
            )

    @property
    def bg_color(self):
        return self.__bg_color

    @bg_color.setter
    def bg_color(self, value):
        if type(value) is str or "FormattingString" in str(type(value)):
            self.__bg_color = value
        else:
            raise base.PglInvalidTypeException(
                f"A Sprixel.bg_color must be a string. {value} is not a string."
            )

    @property
    def fg_color(self):
        return self.__fg_color

    @fg_color.setter
    def fg_color(self, value):
        if type(value) is str or "FormattingString" in str(type(value)):
            self.__fg_color = value
        else:
            raise base.PglInvalidTypeException(
                f"A Sprixel.fg_color must be a string. {value} is not a string."
            )

    def serialize(self):
        """Serialize a Sprixel into a dictionary.

        :returns: The class as a  dictionary
        :rtype: dict

        Example::

            json.dump( sprixel.serialize() )
        """
        return {
            "model": self.model,
            "bg_color": self.bg_color,
            "fg_color": self.fg_color,
            "is_bg_transparent": self.is_bg_transparent,
        }

    @classmethod
    def load(cls, data):
        """
        Create a new Sprixel object based on serialized data.

        :param data: Data loaded from JSON data (deserialized).
        :type data: dict
        :rtype: :class:`Sprixel`

        Example::

            new_sprite = Sprixel.load(json_parsed_data['default_sprixel'])
        """
        sprix = cls(data["model"], data["bg_color"], data["fg_color"])
        sprix.is_bg_transparent = bool(data["is_bg_transparent"])
        return sprix

    @classmethod
    def black_rect(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.BLACK_RECT.
        The difference is that BLACK_RECT is a string and this one is a Sprixel that can
        be manipulated more easily.

        Example::

            sprixel = Sprixel.black_rect()
        """
        return cls(" ", "\x1b[40m")

    @classmethod
    def black_square(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.BLACK_SQUARE.
        The difference is that BLACK_SQUARE is a string and this one is a Sprixel that
        can be manipulated more easily.

        Example::

            sprixel = Sprixel.black_square()
        """
        return cls("  ", "\x1b[40m")

    @classmethod
    def white_rect(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.WHITE_RECT.
        The difference is that WHITE_RECT is a string and this one is a Sprixel that can
        be manipulated more easily.

        Example::

            sprixel = Sprixel.white_rect()
        """
        return cls(" ", "\x1b[47m")

    @classmethod
    def white_square(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.WHITE_SQUARE.
        The difference is that WHITE_SQUARE is a string and this one is a Sprixel that
        can be manipulated more easily.

        Example::

            sprixel = Sprixel.white_square()
        """
        return cls("  ", "\x1b[47m")

    @classmethod
    def red_rect(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.RED_RECT.
        The difference is that RED_RECT is a string and this one is a Sprixel that can
        be manipulated more easily.

        Example::

            sprixel = Sprixel.red_rect()
        """
        return cls(" ", "\x1b[41m")

    @classmethod
    def red_square(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.RED_SQUARE.
        The difference is that RED_SQUARE is a string and this one is a Sprixel that can
        be manipulated more easily.

        Example::

            sprixel = Sprixel.red_square()
        """
        return cls("  ", "\x1b[41m")

    @classmethod
    def green_rect(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.GREEN_RECT.
        The difference is that GREEN_RECT is a string and this one is a Sprixel that can
        be manipulated more easily.

        Example::

            sprixel = Sprixel.green_rect()
        """
        return cls(" ", "\x1b[42m")

    @classmethod
    def green_square(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.GREEN_SQUARE.
        The difference is that GREEN_SQUARE is a string and this one is a Sprixel that
        can be manipulated more easily.

        Example::

            sprixel = Sprixel.green_square()
        """
        return cls("  ", "\x1b[42m")

    @classmethod
    def blue_rect(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.BLUE_RECT.
        The difference is that BLUE_RECT is a string and this one is a Sprixel that can
        be manipulated more easily.

        Example::

            sprixel = Sprixel.blue_rect()
        """
        return cls(" ", "\x1b[44m")

    @classmethod
    def blue_square(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.BLUE_SQUARE.
        The difference is that BLUE_SQUARE is a string and this one is a Sprixel that
        can be manipulated more easily.

        Example::

            sprixel = Sprixel.blue_square()
        """
        return cls("  ", "\x1b[44m")

    @classmethod
    def cyan_rect(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.CYAN_RECT.
        The difference is that CYAN_RECT is a string and this one is a Sprixel that can
        be manipulated more easily.

        Example::

            sprixel = Sprixel.cyan_rect()
        """
        return cls(" ", "\x1b[46m")

    @classmethod
    def cyan_square(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.CYAN_SQUARE.
        The difference is that CYAN_SQUARE is a string and this one is a Sprixel that
        can be manipulated more easily.

        Example::

            sprixel = Sprixel.cyan_square()
        """
        return cls("  ", "\x1b[46m")

    @classmethod
    def magenta_rect(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.MAGENTA_RECT.
        The difference is that MAGENTA_RECT is a string and this one is a Sprixel that
        can be manipulated more easily.

        Example::

            sprixel = Sprixel.magenta_rect()
        """
        return cls(" ", "\x1b[45m")

    @classmethod
    def magenta_square(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.MAGENTA_SQUARE.
        The difference is that MAGENTA_SQUARE is a string and this one is a Sprixel that
        can be manipulated more easily.

        Example::

            sprixel = Sprixel.magenta_square()
        """
        return cls("  ", "\x1b[45m")

    @classmethod
    def yellow_rect(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.YELLOW_RECT.
        The difference is that YELLOW_RECT is a string and this one is a Sprixel that
        can be manipulated more easily.

        .. Note:: Yellow is often rendered as brown.

        Example::

            sprixel = Sprixel.yellow_rect()
        """
        return cls(" ", "\x1b[43m")

    @classmethod
    def yellow_square(cls):
        """
        This classmethod returns a sprixel that is the equivalent of
        pygamelib.assets.graphics.YELLOW_SQUARE.
        The difference is that YELLOW_SQUARE is a string and this one is a Sprixel that
        can be manipulated more easily.

        .. Note:: Yellow is often rendered as brown.

        Example::

            sprixel = Sprixel.yellow_square()
        """
        return cls("  ", "\x1b[43m")


class Sprite(object):
    """
    The Sprite object represent a 2D "image" that can be used to represent any complex
    item.
    Obviously, a sprite in the pygamelib is not really an image, it is a series of
    glyphs (or characters) with colors (foreground and background) information.

    A Sprite object is a 2D array of :class:`Sprixel`.

    If you use the climage python module, you can load the generated result into a
    Sprite through Sprite.load_from_ansi_file().

    :param sprixels: A 2D array of :class:`Sprixel`.
    :type sprixels: list
    :param default_sprixel: A default Sprixel to complete lines that are not long
       enough. By default, it's an empty Sprixel.
    :type default_sprixel: :class:`Sprixel`
    :param parent: The parent object of this Sprite. If it's left to None, the
       :class:`~pygamelib.board_items.BoardComplexItem` constructor takes ownership of
       the sprite.
    :type parent: :class:`~pygamelib.board_items.BoardComplexItem` (suggested)
    :param size: A 2 elements list that represent the width and height ([width, height])
       of the Sprite. It is only needed if you create an empty Sprite. If you load from
       a file or provide an array of sprixels it's obviously calculated automatically.
       Default value: [2, 2].
    :type size: list
    :param name: The name of sprite. If none is given, an UUID will be automatically
       generated.
    :type name: str

    Example::

        void = Sprixel()
        # This represent a panda
        panda_sprite = Sprite(
            sprixels=[
                [void, void, void, void, void, void, void, void],
                [
                    Sprixel.black_rect(),
                    Sprixel.black_rect(),
                    void,
                    void,
                    void,
                    void,
                    Sprixel.black_rect(),
                    Sprixel.black_rect(),
                ],
                [
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                ],
                [
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.black_rect(),
                    Sprixel.black_rect(),
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.black_rect(),
                    Sprixel.black_rect(),
                ],
                [
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.red_rect(),
                    Sprixel.red_rect(),
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                ],
                [
                    void,
                    void,
                    Sprixel.black_rect(),
                    Sprixel.black_rect(),
                    Sprixel.black_rect(),
                    Sprixel.black_rect(),
                    void,
                    void,
                ],
                [
                    void,
                    void,
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.white_rect(),
                    Sprixel.black_rect(),
                    Sprixel.black_rect(),
                ],
                [
                    void,
                    void,
                    Sprixel.black_rect(),
                    Sprixel.black_rect(),
                    void,
                    void,
                    void,
                    void,
                ],
            ],
        )
    """

    def __init__(
        self,
        sprixels=None,
        default_sprixel=Sprixel(),
        parent=None,
        size=[2, 2],
        name=None,
    ):
        super().__init__()
        self.size = size
        self.parent = parent
        self.name = name
        self.default_sprixel = default_sprixel
        if self.name is None or type(self.name) is not str:
            self.name = str(uuid4())
        if sprixels is not None and len(sprixels) > 0:
            self._sprixels = []
            height = 0
            max_width = 0
            for row in sprixels:
                self._sprixels.append([])
                if len(row) > max_width:
                    max_width = len(row)
                width = 0
                for column in row:
                    self._sprixels[height].append(sprixels[height][width])
                    width += 1
                height += 1
            self.size = [max_width, len(self._sprixels)]

        else:
            self.empty()

    def __repr__(self):
        string = []
        for scanline in self._sprixels:
            string.append("".join(map(lambda i: i.__repr__(), scanline)))
        return "\n".join(string)

    def __str__(self):  # pragma: no cover
        return self.__repr__()

    def empty(self):
        """Empty the sprite and fill it with default sprixels.

        Example::

            player_sprite.empty()
        """
        self._sprixels = [
            [self.default_sprixel for i in range(0, self.size[0])]
            for j in range(0, self.size[1])
        ]

    def sprixel(self, row=0, column=None):
        """Return a sprixel at a specific position within the sprite.

        If the column is set to None, the whole row is returned.

        :param row: The row to access within the sprite.
        :type row: int
        :param column: The column to access within the sprite.
        :type column: int

        Example::

            # Return the entire line at row index 2
            scanline = house_sprite.sprixel(2)
            # Return the specific sprixel at sprite internal coordinate 2,3
            house_sprixel = house_sprite.sprixel(2, 3)

        .. WARNING:: For performance consideration sprixel() does not check the size of
           its matrix. This method is called many times during rendering and 2 calls to
           len() in a row are adding up pretty quickly.
           It checks the boundary of the sprite using the cached size. Make sure it is
           up to date!
        """
        if type(row) is not int:
            raise base.PglInvalidTypeException("Sprite.sprixel(): Row is not an int.")
        if row < 0 or row >= self.size[1]:
            raise base.PglException(
                "out_of_sprite_boundaries",
                f"Sprite.sprixel(): Row ({row}) is out of the Sprite boundaries.",
            )
        if column is None:
            return self._sprixels[row]
        else:
            if type(column) is not int:
                raise base.PglInvalidTypeException(
                    "Sprite.sprixel(): Column is not an int."
                )
            if column < 0 or column >= self.size[0]:
                raise base.PglException(
                    "out_of_sprite_boundaries",
                    f"Sprite.sprixel(): Column ({column}) is out of the "
                    "Sprite boundaries.",
                )
            return self._sprixels[row][column]

    def set_sprixel(self, row, column, val):
        """
        Set a specific sprixel in the sprite to the given value.
        :param name: some param
        :type name: str

        Example::

            method()
        """
        if type(row) is not int or type(column) is not int:
            raise base.PglInvalidTypeException(
                "Sprite.set_sprixel(row, column, val) row and column needs to be "
                "integer."
            )
        if row < 0 or row >= self.size[1] or column < 0 or column >= self.size[0]:
            raise base.PglException(
                "out_of_sprite_boundaries",
                f"Sprite.set_sprixel(): ({row},{column}) is out of bound.",
            )
        if not isinstance(val, Sprixel):
            raise base.PglInvalidTypeException(
                "Sprite.set_sprixel(row, column, val) val needs to be a Sprixel"
            )
        self._sprixels[row][column] = val

    @classmethod
    def from_text(cls, text_object):
        """
        Create a Sprite from a :class:`~pygamelib.base.Text` object.

        :param text_object: A text object to transform into Sprite.
        :type text_object: :class:`~pygamelib.base.Text`

        Example::

            # The Text object allow for easy manipulation of text
            village_name = base.Text('khukdale',fg_red, bg_green)
            # It can be converted into a Sprite to be displayed on the Board
            village_sign = board_items.Tile(sprite=Sprite.from_text(village_name))
            # And can be used as formatted text
            notifications.push( f'You enter the dreaded village of {village_name}' )
        """
        sprixels = []
        style = ""
        max_width = 0
        for line in text_object.text.splitlines():
            sprixels.append([])
            if len(line) > max_width:
                max_width = len(line)
            for char in line:
                sprixels[-1].append(
                    Sprixel(
                        style + char + base.Style.RESET_ALL,
                        text_object.bg_color,
                        text_object.fg_color,
                    )
                )
        return cls(sprixels=sprixels)

    @classmethod
    def load_from_ansi_file(cls, filename, default_sprixel=None):
        """Load an ANSI encoded file into a Sprite object.

        This class method can load a file produced by the climage python module and
        load it into a Sprite class. Each character is properly decoded into a
        :class:`Sprixel` with model, background and foreground colors.

        A Sprite is rectangular (at least for the moment), so in case the file is
        not shaped as a rectangle, this method automatically fills the void with a
        default sprixel (to make sure all lines in the sprite have the same length).
        By default, it fills the table with None "values" but you can specify a default
        sprixel.

        The reasons the default sprixel is set to None is because None values in a
        sprite are not translated into a component in
        :class:`~pygamelib.board_items.BoardComplexItem` (i.e no sub item is generated).

        :param filename: The path to a file to load.
        :type filename: str
        :param default_sprixel: The default Sprixel to fill a non rectangular shaped
           sprite.
        :type default_sprixel: None | :class:`Sprixel`

        Example::

            player_sprite = gfx_core.Sprite.load_from_ansi_file('gfx/models/player.ans')
        """
        if default_sprixel is None:
            default_sprixel = Sprixel(" ", "", "")
        new_sprite = cls(default_sprixel=default_sprixel)
        with open(filename, "r") as sprite_file:
            sprixels_list = []
            height = 0
            max_width = 0
            while True:
                line = sprite_file.readline()
                if not line:
                    break
                width = 0
                sprixels_list.append([])
                for s in line.rstrip().split("▄"):
                    if s != "\x1b[0m":
                        sprixels_list[height].append(Sprixel.from_ansi(f"{s}▄\x1b[0m"))
                        width += 1
                if width > max_width:
                    max_width = width
                height += 1
            for row in range(0, len(sprixels_list)):
                if len(sprixels_list[row]) < max_width:
                    for column in range(len(sprixels_list[row]), max_width):
                        sprixels_list[row].append(default_sprixel)
            new_sprite._sprixels = sprixels_list
            new_sprite.size = [max_width, height]
        return new_sprite

    def flip_horizontally(self):
        """Flip the sprite horizontally.

        This method performs a symmetry versus the vertical axis.

        At the moment, glyph are not inverted. Only the position of the sprixels.

        The flipped sprite is returned (original sprite is not modified).

        :rtype: :class:`Sprite`

        Example::

            reflection_sprite = player_sprite.flip_horizontally()
        """
        new_sprite = Sprite(
            size=self.size,
            sprixels=None,
            default_sprixel=self.default_sprixel,
            parent=self.parent,
        )
        nc = 0
        # Flipping horizontally is just a symmetry vs a vertical axis
        for col in range(self.size[0] - 1, -1, -1):
            for row in range(0, self.size[1]):
                new_sprite.set_sprixel(row, nc, self._sprixels[row][col])
            nc += 1
        return new_sprite

    def flip_vertically(self):
        """Flip the sprite vertically (i.e upside/down).

        At the moment, glyph are not inverted. Only the position of the sprixels.
        There is one exception however, as climage uses the '▄' utf8 glyph as a marker,
        that specific glyph is inverted to '▀' and vice versa.

        The flipped sprite is returned (original sprite is not modified).

        :rtype: :class:`Sprite`

        Example::

            reflection_sprite = player_sprite.flip_vertically()
        """
        new_sprite = Sprite(
            size=self.size,
            sprixels=None,
            default_sprixel=self.default_sprixel,
            parent=self.parent,
        )
        nr = 0
        for row in range(self.size[1] - 1, -1, -1):
            for col in range(0, self.size[0]):
                new_sprix = self._sprixels[row][col]
                if new_sprix.model == graphics.Blocks.LOWER_HALF_BLOCK:
                    new_sprix.model = graphics.Blocks.UPPER_HALF_BLOCK
                elif new_sprix.model == graphics.Blocks.UPPER_HALF_BLOCK:
                    new_sprix.model = graphics.Blocks.LOWER_HALF_BLOCK
                new_sprite.set_sprixel(nr, col, new_sprix)
            nr += 1
        return new_sprite

    def calculate_size(self):
        """
        Calculate the size of the sprite and update the size variable.

        The size is immediately returned.

        It is done separately for concerns about performances of doing that everytime
        the size is requested.

        :rtype: list

        Example::

            spr_size = spr.calculate_size()
            if spr_size != spr.size:
                raise PglException(
                            'perturbation_in_the_Force',
                            'Something is very wrong with the sprite!'
                        )
        """
        # Warning: recalculate the size of the Sprite, it is much faster
        # although not safe to use self.size
        height = 0
        max_width = 0
        for row in self._sprixels:
            width = 0
            for col in row:
                width += 1
            if width > max_width:
                max_width = width
            height += 1
        self.size = [max_width, height]
        return self.size

    def set_transparency(self, state):
        """This method enable transparent background to all the sprite's sprixels.

        :param state: a boolean to enable or disable background transparency
        :type name: bool

        Example::

            player_sprite.set_transparency(True)

        .. WARNING:: This set background transparency on all sprixels, make sure you are
           not using background colors as part of your sprite before doing that.
           It can also be used as a game/rendering mechanic. Just make sure you know
           what you do.
           As a reminder, by default, sprixels with no background have transparent
           background enable.
        """
        for line in self._sprixels:
            for s in line:
                s.is_bg_transparent = state

    def serialize(self):
        """Serialize a Sprite into a dictionary.

        :returns: The class as a  dictionary
        :rtype: dict

        Example::

            json.dump( sprite.serialize() )
        """
        ret_dict = {
            "size": self.size,
            "name": self.name,
            "default_sprixel": self.default_sprixel.serialize(),
            "sprixels": [],
        }
        for row in range(0, self.size[1]):
            tmp = []
            for column in range(0, self.size[0]):
                tmp.append(self.sprixel(row, column).serialize())

            ret_dict["sprixels"].append(tmp)

        return ret_dict

    @classmethod
    def load(cls, data):
        """
        Create a new Sprite object based on serialized data.

        :param data: Data loaded from a JSON sprite file (deserialized).
        :type data: dict
        :rtype: :class:`Sprite`

        Example::

            new_sprite = Sprite.load(json_parsed_data)
        """
        sprixels = []
        try:
            for row in range(0, int(data["size"][1])):
                tmp = []
                for column in range(0, int(data["size"][0])):
                    tmp.append(Sprixel.load(data["sprixels"][row][column]))
                sprixels.append(tmp)
            new_sprite = cls(
                name=data["name"],
                default_sprixel=Sprixel.load(data["default_sprixel"]),
                sprixels=sprixels,
            )
            return new_sprite
        except IndexError:
            raise base.PglException(
                "invalid_sprite_size",
                "Sprite.load(data): sprixels array size is different from sprite "
                "computed size.",
            )


class SpriteCollection(UserDict):
    """
    SpriteCollection is a dictionnary class that derives collections.UserDict.

    Its main goal is to provide an easy to use object to load and save sprite files.
    On top of traditional dict method, it provides the following capabilities:

     - loading and writing from and to JSON files,
     - data serialization,
     - shortcut to add sprites to the dictionnary.

    A SpriteCollection is an unordered indexed list of Sprites (i.e a dictionnary).

    Sprites are indexed by their names in that collection.

    Example::

       # Load a sprite file
       sprites_village1 = SpriteCollection.load_json_file('gfx/village1.spr')
       # display the Sprites with their name
       for sprite_name in sprites_village1:
           print(f'{sprite_name}:\\n{sprites_village1[sprite_name]}')
       # Add an empty sprite with name 'house_placeholder'
       sprites_village1.add( Sprite(name='house_placeholder') )
       # This is absolutely equivalent to:
       sprites_village1['house_placeholder'] = Sprite(name='house_placeholder')
       # And now rewrite the sprite file with the new placeholder house
       sprites_village1.to_json_file('gfx/village1.spr')

    """

    def __init__(self, data=dict()):
        UserDict.__init__(self, data)

    @classmethod
    def load(cls, data):
        """
        Load serialized data and return a new SpriteCollection object.

        :param data: Serialized data that need to be expanded into objects.
        :type data: str
        :returns: A new SpriteCollection object.
        :rtype: :class:`SpriteCollection`

        Example::

            sprites_village1 = SpriteCollection.load(
                sprites_village_template.serialize()
            )
        """
        if "sprites_count" in data.keys() and data["sprites_count"] is not None:
            if (
                "sprites" in data.keys()
                and data["sprites"] is not None
                and type(data["sprites"]) is dict
            ):
                sc = cls()
                if data["sprites_count"] == len(data["sprites"]):
                    for spr_name in data["sprites"]:
                        sc.data[spr_name] = Sprite.load(data["sprites"][spr_name])
                    return sc
                else:
                    raise base.PglException(
                        "corrupted_sprite_data",
                        "SpriteCollection.load(data): sprites count is different from "
                        "actual number of sprites.",
                    )
        else:
            raise base.PglException(
                "invalid_sprite_data",
                "SpriteCollection.load(data): Invalid sprite data, sprites_count is "
                "missing.",
            )

    @staticmethod
    def load_json_file(filename):
        """
        Load a JSON sprite file into a new SpriteCollection object.

        :param filename: The complete path (relative or absolute) to the sprite file.
        :type filename: str
        :returns: A new SpriteCollection object.
        :rtype: :class:`SpriteCollection`

        Example::

            sprites_village1 = SpriteCollection.load_json_file('gfx/village1.spr')
        """
        with open(filename) as sprites_file:
            sprites_data = json.load(sprites_file)
            return SpriteCollection.load(sprites_data)

    def serialize(self):
        """
        Return a serialized version of the SpriteCollection. The serialized data can be
        pass to the JSON module to export.

        :returns: The SpriteCollection object serialized as a dictionnary.
        :rtype: dict

        Example::

            data = sprites_village1.serialize()
        """
        ret_data = dict()
        ret_data["sprites_count"] = len(self.data)
        ret_data["sprites"] = dict()
        for spr_name in self.data:
            ret_data["sprites"][spr_name] = self.data[spr_name].serialize()
        return ret_data

    def to_json_file(self, filename):
        """
        Export the SpriteCollection object in JSON and writes it on the disk.

        :param filename: The complete path (relative or absolute) to the sprite file to
           write.
        :type filename: str

        Example::

            sprites_village1.to_json_file('gfx/village1.spr')
        """
        with open(filename, "w") as file:
            json.dump(self.serialize(), file)

    def add(self, sprite):
        """
        Add a Sprite to the collection. This method is simply a shortcut to the usual
        dictionnary affectation. The collection requires the name of the Sprite to be
        the key. That method does that automatically.

        :param sprite: A Sprite object to add to the collection.
        :type sprite: :class:`Sprite`

        .. WARNING:: As SpriteCollection index Sprites by their name if you change the
           Sprite's name *after* adding it to the collection you will need to manually
           update the keys.

        Example::

            sprites_village1 = SpriteCollection.load_json_file('gfx/village1.spr')
            new_village = SpriteCollection()
            new_village.add( copy.deepcopy( sprites_village1.get('bakery') ) )
            print( new_village['bakery'] )
        """
        if isinstance(sprite, Sprite):
            self.data[sprite.name] = sprite
        else:
            raise base.PglInvalidTypeException(
                "SpriteCollection.add(sprite): require a Sprite object."
            )


class Animation(object):
    """
    The Animation class is used to give the ability to have more than one model
    for a BoardItem. A BoardItem can have an animation and all of them that
    are available to the Game object can be animated through
    Game.animate_items(lvl_number).
    To benefit from that, BoardItem.animation must be set explicitely.
    An animation is controlled via the same state system than the Actuators.

    The frames are all stored in a list called frames, that you can access
    through Animation.frames.

    :param display_time: The time each frame is displayed
    :type display_time: float
    :param auto_replay: controls the auto replay of the animation, if false
        once the animation is played it stays on the last
        frame of the animation.
    :type auto_replay: bool
    :param frames: an array of "frames" (string, sprixel or sprite)
    :type frames: array[str|Sprixel|Sprite]
    :param animated_object: The object to animate. This parameter is deprecated.
        Please use parent instead. It is only kept for backward compatibility.
        The parent parameter always takes precedence over this one.
    :type animated_object: :class:`~pygamelib.board_items.BoardItem`
    :param parent: The parent object. It is also the object to animate.
        Important: We cannot animate anything else that BoardItems and subclasses.
    :type parent: :class:`~pygamelib.board_items.BoardItem`
    :param refresh_screen: The callback function that controls the redrawing of
        the screen. This function reference should come from the main game.
    :type refresh_screen: function

    Example ::

        def redraw_screen(game_object):
            game_object.clear_screen()
            game_object.display_board()

        item = BoardItem(model=Sprite.ALIEN, name='Friendly Alien')
        # By default BoardItem does not have any animation, we have to
        # explicitely create one
        item.animation = Animation(display_time=0.1, parent=item,
                                   refresh_screen=redraw_screen)
    """

    def __init__(
        self,
        display_time=0.05,
        auto_replay=True,
        frames=None,
        animated_object=None,
        refresh_screen=None,
        initial_index=None,
        parent=None,
    ):
        self.state = constants.RUNNING
        self.display_time = display_time
        self.auto_replay = auto_replay
        if frames is None:
            frames = []
        elif isinstance(frames, SpriteCollection):
            nf = []
            for sn in frames:
                nf.append(frames[sn])
            frames = nf

        self.frames = frames
        if initial_index is None:
            self._frame_index = 0
            self._initial_index = 0
        else:
            self._frame_index = initial_index
            self._initial_index = initial_index
        if animated_object is not None and parent is None:
            self.parent = animated_object
            self.animated_object = animated_object
        elif parent is not None:
            self.parent = parent
            self.animated_object = parent
        self.refresh_screen = refresh_screen
        self.__dtanimate = 0.0

    @property
    def dtanimate(self):
        return self.__dtanimate

    @dtanimate.setter
    def dtanimate(self, value):
        if type(value) is float or type(value) is int:
            self.__dtanimate = value
        else:
            raise base.PglInvalidTypeException(
                "Animation.dtanimate(value): value needs to be an int or float."
            )

    def start(self):
        """Set the animation state to constants.RUNNING.

        If the animation state is not constants.RUNNING, animation's next_frame()
        function return the last frame returned.

        Example::

            item.animation.start()
        """
        self.state = constants.RUNNING

    def pause(self):
        """Set the animation state to PAUSED.

        Example::

            item.animation.pause()
        """
        self.state = constants.PAUSED

    def stop(self):
        """Set the animation state to STOPPED.

        Example::

            item.animation.stop()
        """
        self.state = constants.STOPPED

    def add_frame(self, frame):
        """Add a frame to the animation.

        The frame has to be a string (that includes sprites from the Sprite
        module and squares from the Utils module).

        Raise an exception if frame is not a string.

        :param frame: The frame to add to the animation.
        :type frame: str
        :raise: :class:`pygamelib.base.PglInvalidTypeException`

        Example::

            item.animation.add_frame(Sprite.ALIEN)
            item.animation.add_frame(Sprite.ALIEN_MONSTER)
        """
        if (
            type(frame) is str
            or isinstance(frame, Sprixel)
            or isinstance(frame, Sprite)
        ):
            self.frames.append(frame)
        else:
            raise base.PglInvalidTypeException(
                'The "frame" parameter must be a string, Sprixel or Sprite.'
            )

    def search_frame(self, frame):
        """Search a frame in the animation.

        That method is returning the index of the first occurrence of "frame".

        Raise an exception if frame is not a string.

        :param frame: The frame to find.
        :type frame: str
        :rtype: int
        :raise: :class:`~pygamelib.base.PglInvalidTypeException`

        Example::

            item.animation.remove_frame(
                item.animation.search_frame(Sprite.ALIEN_MONSTER)
            )

        """
        if (
            type(frame) is str
            or isinstance(frame, Sprixel)
            or isinstance(frame, Sprite)
        ):
            return self.frames.index(frame)
        else:
            raise base.PglInvalidTypeException(
                'The "frame" parameter must be a string.'
            )

    def remove_frame(self, index):
        """Remove a frame from the animation.

        That method remove the frame at the specified index and return it
        if it exists.

        If the index is out of bound an exception is raised.
        If the index is not an int an exception is raised.

        :param index: The index of the frame to remove.
        :type index: int
        :rtype: str
        :raise: IndexError, PglInvalidTypeException

        Example::

            item.animation.remove_frame( item.animation.search_frame(
                Sprite.ALIEN_MONSTER)
            )

        """
        if type(index) is not int:
            raise base.PglInvalidTypeException('The "index" parameter must be an int.')
        if index <= self._frame_index and self._frame_index > 0:
            self._frame_index -= 1
        return self.frames.pop(index)

    def reset(self):
        """Reset the Animation to the first frame.

        Example::

            item.animation.reset()
        """
        self._frame_index = self._initial_index

    def current_frame(self):
        """Return the current frame.

        Example::

            item.model = item.animation.current_frame()
        """
        return self.frames[self._frame_index]

    def next_frame(self):
        """Update the parent's model, sprixel or sprite with the next frame of the
        animation.

        That method takes care of automatically replaying the animation if the
        last frame is reached if the state is constants.RUNNING.

        If the the state is PAUSED it still update the parent.model
        and returning the current frame. It does NOT actually go to next frame.

        If parent is not a sub class of
        :class:`~pygamelib.board_items.BoardItem` an exception is raised.

        :raise: :class:`~pygamelib.base.PglInvalidTypeException`

        Example::

            item.animation.next_frame()

        .. WARNING:: If you use Sprites as frames, you need to make sure your Animation
            is attached to a :class:`~pygamelib.board_items.BoardComplexItem`.

        """
        if not isinstance(self.parent, board_items.BoardItem):
            raise base.PglInvalidTypeException(
                "The parent needs to be a sub class of BoardItem."
            )
        if self.state == constants.STOPPED:
            return
        elif self.state == constants.RUNNING:
            self._frame_index += 1
            if self._frame_index >= len(self.frames):
                if self.auto_replay:
                    self.reset()
                else:
                    self._frame_index = len(self.frames) - 1
        if type(self.frames[self._frame_index]) is str:
            self.parent.model = self.frames[self._frame_index]
        elif isinstance(self.frames[self._frame_index], Sprixel):
            self.parent.sprixel = self.frames[self._frame_index]
        elif isinstance(self.frames[self._frame_index], Sprite):
            self.parent.sprite = self.frames[self._frame_index]
            self.parent.update_sprite()
        else:
            raise base.PglInvalidTypeException(
                "Animation.next_frame(): the frame is neither a string, a sprixel nor a"
                " sprite."
            )
        return self.frames[self._frame_index]

    def play_all(self):
        """Play the entire animation once.

        That method plays the entire animation only once, there is no auto
        replay as it blocks the game (for the moment).

        If the the state is PAUSED or STOPPED, the animation does not play and
        the method return False.

        If parent is not a sub class of
        :class:`~pygamelib.board_items.BoardItem` an exception is raised.

        If screen_refresh is not defined or is not a function an exception
        is raised.

        :raise: :class:`~pygamelib.base.PglInvalidTypeException`

        Example::

            item.animation.play_all()
        """
        if self.state == constants.PAUSED or self.state == constants.STOPPED:
            return False
        if self.refresh_screen is None or not callable(self.refresh_screen):
            raise base.PglInvalidTypeException(
                "The refresh_screen parameter needs to be a callback "
                "function reference."
            )
        if not isinstance(self.parent, board_items.BoardItem):
            raise base.PglInvalidTypeException(
                "The parent needs to be a sub class of BoardItem."
            )
        self.reset()
        previous_time = time.time()
        ctrl = 0
        while ctrl < len(self.frames):
            if self.dtanimate >= self.display_time:
                # Dirty but that's a current limitation: to restore stuff on the board's
                # overlapped matrix, we need to either move or replace an item after
                # updating the sprite. This is only for sprites that have null items but
                # we don't want to let any one slip.
                # Also: this is convoluted...
                if (
                    isinstance(self.parent, board_items.BoardComplexItem)
                    and self.parent.parent is not None
                    and (
                        isinstance(self.parent.parent, board_items.engine.Board)
                        or isinstance(self.parent.parent, board_items.engine.Game)
                    )
                ):
                    b = None
                    if isinstance(self.parent.parent, board_items.engine.Board):
                        b = self.parent.parent
                    else:
                        b = self.parent.parent.current_board()
                    pos = self.parent.pos
                    # We have to think that someone could try to animate the player
                    # while not on the current board.
                    try:
                        b.remove_item(self.parent)
                    except Exception:
                        return
                    self.next_frame()
                    b.place_item(self.parent, pos[0], pos[1])
                else:
                    self.next_frame()
                self.refresh_screen()
                ctrl += 1
                self.dtanimate = 0
            self.dtanimate += time.time() - previous_time
            previous_time = time.time()
        return True



"""This module contains the basic board items classes.

.. autosummary::
   :toctree: .

   BoardItem
   BoardItemVoid
   BoardComplexItem
   BoardItemComplexComponent
   Movable
   Projectile
   Immovable
   Actionable
   Character
   Player
   ComplexPlayer
   NPC
   ComplexNPC
   TextItem
   Wall
   ComplexWall
   Treasure
   ComplexTreasure
   Door
   ComplexDoor
   GenericStructure
   GenericActionableStructure
   Tile
"""
from pygamelib import engine
from pygamelib import base
from pygamelib import constants
from pygamelib.gfx import core
from pygamelib import actuators


class BoardItem:
    """
    Base class for any item that will be placed on a Board.

    :param type: A type you want to give your item. It can be any string. You can then
        use the type for sorting or grouping for example.
    :type type: str
    :param name: A name for this item. For identification purpose.
    :type name: str
    :param pos: the position of this item. When the item is managed by the Board and
        Game engine this member hold the last updated position of the item. It is not
        updated if you manually move the item. It must be an array of
        2 integers [row,column]
    :type pos: array
    :param model: The model to use to display this item on the Board. Be mindful of the
        space it will require. Default value is '*'.
    :type model: str
    :param parent: The parent object of the board item. Usually a Board or Game object.

    .. important:: Starting with version 1.2.0 and introduction of complex items,
       BoardItems have a size. That size **CANNOT** be set. It is always 1x1.
       This is because a BoardItem always takes 1 cell, regardless of its actual number
       of characters. Python does not really provide a way to prevent changing that
       member but if you do, you'll break rendering. You have been warned.
    """

    def __init__(self, **kwargs):
        self.name = "Board item"
        self.type = "item"
        self.pos = [None, None]
        self.model = "*"
        self.animation = None
        self.parent = None
        self.sprixel = None
        self.size = [1, 1]
        # Setting class parameters
        for item in ["name", "type", "pos", "model", "parent", "sprixel"]:
            if item in kwargs:
                setattr(self, item, kwargs[item])

    def __str__(self):
        if self.sprixel is not None:
            return self.sprixel.__repr__()
        else:
            return self.model

    def __repr__(self):  # pragma: no cover
        return self.__str__()

    def display(self):
        """
        Print the model WITHOUT carriage return.
        """
        print(self.model, end="")

    def debug_info(self):
        """
        Return a string with the list of the attributes and their current value.

        :rtype: str
        """
        string = "attrs: \n"
        for key in vars(self):
            if type(getattr(self, key)) is list:
                string += (
                    f"'{key}' = '"
                    + "".join(str(e) + " " for e in getattr(self, key))
                    + "'\n"
                )
            else:
                string += f"'{key}' = '{getattr(self, key)}'\n"
        return string

    def store_position(self, row, column):
        """Store the BoardItem position for self access.

        The stored position is used for consistency and quick access to the self
        postion. It is a redundant information and might not be synchronized.

        :param row: the row of the item in the :class:`~pygamelib.engine.Board`.
        :type row: int
        :param column: the column of the item in the :class:`~pygamelib.engine.Board`.
        :type column: int

        Example::

            item.store_position(3,4)
        """
        self.pos = [row, column]

    def position_as_vector(self):
        """Returns the current item position as a Vector2D

        :returns: The position as a 2D vector
        :rtype: :class:`~pygamelib.base.Vector2D`

        Example::

            gravity = Vector2D(9.81, 0)
            next_position = item.position_as_vector() + gravity.unit()
        """
        return base.Vector2D(self.pos[0], self.pos[1])

    @property
    def row(self):
        """Convenience method to get the current stored row of the item.

        This is absolutely equivalent to access to item.pos[0].

        :return: The row coordinate
        :rtype: int

        Example::

            if item.row != item.pos[0]:
                print('Something extremely unlikely just happened...')
        """
        return self.pos[0]

    @property
    def column(self):
        """Convenience method to get the current stored column of the item.

        This is absolutely equivalent to access to item.pos[1].

        :return: The column coordinate
        :rtype: int

        Example::

            if item.column != item.pos[1]:
                print('Something extremely unlikely just happened...')
        """
        return self.pos[1]

    @property
    def width(self):
        """Convenience method to get the width of the item.

        This is absolutely equivalent to access to item.size[0].

        :return: The width
        :rtype: int

        Example::

            if item.width > board.width:
                print('The item is too big for the board.')
        """
        return self.size[0]

    @property
    def height(self):
        """Convenience method to get the height of the item.

        This is absolutely equivalent to access to item.size[1].

        :return: The height
        :rtype: int

        Example::

            if item.height > board.height:
                print('The item is too big for the board.')
        """
        return self.size[1]

    def collides_with(self, other):
        """Tells if this item collides with another item.

        :param other: The item you want to check for collision.
        :type other: :class:`~pygamelib.board_items.BoardItem`
        :rtype: bool

        Example::

            if projectile.collides_with(game.player):
                game.player.hp -= 5
        """
        if isinstance(other, BoardItem):
            return base.Math.intersect(
                self.pos[0],
                self.pos[1],
                self.size[0],
                self.size[1],
                other.pos[0],
                other.pos[1],
                other.size[0],
                other.size[1],
            )
        else:
            raise base.PglInvalidTypeException(
                "BoardItem.collides_with require a BoardItem as parameter."
            )

    def distance_to(self, other):
        """Calculates the distance with an item.

        :param other: The item you want to calculate the distance to.
        :type other: :class:`~pygamelib.board_items.BoardItem`
        :return: The distance between this item and the other.
        :rtype: float

        Example::

            if npc.distance_to(game.player) <= 2.0:
                npc.seek_and_destroy = True
        """
        if isinstance(other, BoardItem):
            return base.Math.distance(
                self.pos[0], self.pos[1], other.pos[0], other.pos[1],
            )
        else:
            raise base.PglInvalidTypeException(
                "BoardItem.distance_to require a BoardItem as parameter."
            )

    def can_move(self):
        """
        This is a virtual method that must be implemented in deriving classes.
        This method has to return True or False.
        This represent the capacity for a BoardItem to be moved by the Board.
        """
        raise NotImplementedError()

    def pickable(self):
        """
        This is a virtual method that must be implemented in deriving class.
        This method has to return True or False.
        This represent the capacity for a BoardItem to be pick-up by player or NPC.
        """
        raise NotImplementedError()

    def overlappable(self):
        """
        This is a virtual method that must be implemented in deriving class.
        This method has to return True or False.
        This represent the capacity for a BoardItem to be overlapped by another
        BoardItem.
        """
        raise NotImplementedError()

    def inventory_space(self):
        """
        This is a virtual method that must be implemented in deriving class.
        This method has to return an integer.
        This represent the size of the BoardItem for the
        :class:`~pygamelib.engine.Inventory`. It is used for example to evaluate
        the space taken in the inventory.

        .. important:: That abstract function was called size() before version 1.2.0.
           As it was exclusively used for inventory space management, it as been
           renamed. Particularly because now items do have a need for a size.
        """
        raise NotImplementedError()


class BoardItemVoid(BoardItem):
    """
    A class that represent a void cell.
    """

    def __init__(self, **kwargs):
        BoardItem.__init__(self, **kwargs)
        self.name = "void_cell"

    def pickable(self):
        """
        A BoardItemVoid is not pickable, therefor this method return false.

        :return: False
        """
        return False

    def overlappable(self):
        """
        A BoardItemVoid is obviously overlappable (so player and NPC can walk over).

        :return: True
        """
        return True


class BoardItemComplexComponent(BoardItem):
    """The default component of a complex item.

    It is literrally just a BoardItem but is subclassed for easier identification.

    It is however scanning its parent for the item's basic properties (overlappable,
    restorable, etc.)

    A component can never be pickable by itself.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if hasattr(self, "parent") and self.parent is not None:
            if hasattr(self.parent, "restorable") and callable(self.parent.restorable):
                self.__is_restorable = self.parent.restorable()
            else:
                self.__is_restorable = False
            if hasattr(self.parent, "overlappable") and callable(
                self.parent.overlappable
            ):
                self.__is_overlappable = self.parent.overlappable()
            else:
                self.__is_overlappable = False
            if hasattr(self.parent, "can_move") and callable(self.parent.can_move):
                self.__can_move = self.parent.can_move()
            else:
                self.__can_move = False
        else:
            self.__is_restorable = False
            self.__is_overlappable = False
            self.__can_move = False
        self.__is_pickable = False

    def restorable(self):
        """
        Returns True if the item is restorable, False otherwise.

        Example::

            if item.item(4,5).restorable():
                print('The item is restorable')
        """
        return self.__is_restorable

    def overlappable(self):
        """
        Returns True if the item is overlappable, False otherwise.

        Example::

            if item.item(4,5).overlappable():
                print('The item is overlappable')
        """
        return self.__is_overlappable

    def can_move(self):
        """
        Returns True if the item can move, False otherwise.

        Example::

            if item.item(4,5).can_move():
                print('The item can move')
        """
        return self.__can_move

    def pickable(self):
        """
        Returns True if the item is pickable, False otherwise.

        Example::

            if item.item(4,5).pickable():
                print('The item is pickable')
        """
        return self.__is_pickable


class BoardComplexItem(BoardItem):
    """
    .. versionadded:: 1.2.0

    A BoardComplexItem is the base item for multi cells elements. It inherits from
    :class:`BoardItem` and accepts all its parameters.

    The main difference is that a complex item can use
    :class:`~pygamelib.gfx.core.Sprite` as representation.

    You can see a complex item as a collection of other items that are ruled by the
    same laws. They behave as one but a complex item is actually made of complex
    components. At first it is not important but you may want to exploit that as a
    feature for your game.

    On top of :class:`BoardItem` the constructor accepts the following parameters:

    :param sprite: A sprite representing the item.
    :type sprite: :class:`~pygamelib.gfx.core.Sprite`
    :param size: The size of the item. It impact movement and collision detection
        amongst other things. If it is left empty the Sprite size is used. If no sprite
        is given to the constructor the default size is 2x2.
    :type size: array[int]
    :null_sprixel: The null_sprixel is a bit of a special parameter: during construction
        a null sprixel is replaced by a BoardItemVoid. This is a trick to show the
        background (i.e transparency). A sprixel can take the color of the background
        but a complex item with a null_sprixel that correspond to transparent zone of a
        sprite will really be transparent and show the background.
    :null_sprixel: :class:`~pygamelib.gfx.core.Sprixel`
    :param base_item_type: the building block of the complex item. The complex item is
        built from a 2D array of base items.
    :type base_item_type: :class:`BoardItemComplexComponent`

    """

    def __init__(self, **kwargs):
        self.__kwargs = kwargs
        self.name = "Board Multi Item"
        self.type = "multi_item"
        super().__init__(**kwargs)
        self.sprite = core.Sprite()
        self.null_sprixel = None
        self.size = None
        self._item_matrix = []
        # Not sure about that one
        self.hit_box = []
        self.base_item_type = BoardItemComplexComponent
        for item in ["sprite", "size", "null_sprixel", "base_item_type"]:
            if item in kwargs:
                setattr(self, item, kwargs[item])
        if self.size is None:
            self.size = self.sprite.size
        if isinstance(self.sprite, core.Sprite) and self.sprite.parent is None:
            self.sprite.parent = self
        self.update_sprite()

    def __repr__(self):  # pragma: no cover
        return self.sprite.__repr__()

    def __str__(self):  # pragma: no cover
        return self.sprite.__str__()

    def update_sprite(self):
        """
        Update the complex item with the current sprite.
        This method needs to be called everytime the sprite is changed.

        Example::

            item = BoardComplexItem(sprite=position_idle)
            for s in [walk_1, walk_2, walk_3, walk_4]:
                item.sprite = s
                item.update_sprite()
                board.move(item, constants.RIGHT, 1)
                time.sleep(0.2)
        """
        self._item_matrix = []
        # Update sprite size.
        self.sprite.calculate_size()
        for row in range(0, self.sprite.size[1]):
            self._item_matrix.append([])
            for col in range(0, self.sprite.size[0]):
                if (
                    self.null_sprixel is not None
                    and self.sprite.sprixel(row, col) == self.null_sprixel
                ):
                    self._item_matrix[row].append(BoardItemVoid())
                else:
                    self._item_matrix[row].append(self.base_item_type(**self.__kwargs))
                    self._item_matrix[row][col].name = f"{self.name}_{row}_{col}"
                    self._item_matrix[row][col].model = self.sprite.sprixel(
                        row, col
                    ).model
                    self._item_matrix[row][col].sprixel = self.sprite.sprixel(row, col)
                    self._item_matrix[row][col].parent = self
        self.size = self.sprite.size

    def item(self, row, column):
        """
        Return the item at the row, column position if it is within the item's
        boundaries.

        :rtype: pygamelib.board_items.BoardItem

        :raise PglOutOfBoardBoundException: if row or column are
            out of bound.
        """
        if row < self.size[1] and column < self.size[0]:
            return self._item_matrix[row][column]
        else:
            raise base.PglOutOfItemBoundException(
                (
                    f"There is no item at coordinates [{row},{column}] "
                    "because it's out of the board multi item boundaries "
                    f"({self.size[0]}x{self.size[1]})."
                )
            )


class Movable(BoardItem):
    """A class representing BoardItem capable of movements.

    Movable subclasses :class:`BoardItem`.

    :param step: the amount of cell a movable can cross in one turn. Default value: 1.
    :type step: int
    :param step_vertical: the amount of cell a movable can vertically cross in one turn.
       Default value: step value.
    :type step_vertical: int
    :param step_horizontal: the amount of cell a movable can horizontally cross in one
       turn. Default value: step value.
    :type step_horizontal: int
    :param movement_speed: The time (in seconds) between 2 movements of a Movable. It is
       used by all the Game's actuation methods to enforce move speed of NPC and
       projectiles.
    :type movement_speed: int|float

    The movement_speed parameter is only used when the Game is configured with MODE_RT.
    Additionally the dtmove property is used to accumulate time between frames. It is
    entirely managed by the Game object and most of the time you shouldn't mess up with
    it. Unless you want to manage movements by yourself. If so, have fun! That's the
    point of the pygamelib to let you do whatever you like.

    This class derive BoardItem and describe an object that can move or be
    moved (like a player or NPC).
    Thus this class implements BoardItem.can_move().
    However it does not implement BoardItem.pickable() or
    BoardItem.overlappable()
    """

    def __init__(self, **kwargs):
        BoardItem.__init__(self, **kwargs)
        # We probably need the item to store its own velocity at some point
        # self.velocity = base.Vector2d(0,0)
        # Set default values
        for s in ["step", "step_vertical", "step_horizontal"]:
            if s not in kwargs.keys():
                # self.step = 1
                self.__setattr__(s, 1)
            else:
                # self.step = kwargs["step"]
                self.__setattr__(s, kwargs[s])
        # TODO: That's initial thought for physic i the pygamelib. For future reference.
        # if "velocity" in kwargs.keys():
        #     self.velocity = kwargs["velocity"]
        # else:
        #     self.velocity = base.Vector2D()
        # if "ignore_physic" in kwargs.keys():
        #     self.ignore_physic = kwargs["ignore_physic"]
        # else:
        #     self.ignore_physic = False
        # Now if only step is set and it's not 1, set the correct value for the 2 others
        if "step_vertical" not in kwargs.keys() and self.step != 1:
            self.step_vertical = self.step

        if "step_horizontal" not in kwargs.keys() and self.step != 1:
            self.step_horizontal = self.step
        if "movement_speed" not in kwargs.keys():
            self.movement_speed = 1.0
        else:
            self.movement_speed = kwargs["movement_speed"]
        self.__dtmove = 0.0
        self.__can_move = True

    @property
    def dtmove(self):
        return self.__dtmove

    @dtmove.setter
    def dtmove(self, value):
        if type(value) is float or type(value) is int:
            self.__dtmove = value
        else:
            raise base.PglInvalidTypeException(
                "Movable.dtmove(value): value needs to be an int or float."
            )

    def can_move(self):
        """
        Movable implements can_move().

        :return: True
        :rtype: Boolean
        """
        return self.__can_move

    def has_inventory(self):
        """
        This is a virtual method that must be implemented in deriving class.
        This method has to return True or False.
        This represent the capacity for a Movable to have an inventory.
        """
        raise NotImplementedError()


class Projectile(Movable):
    """A class representing a projectile type board item.
    That class can be sub-classed to represent all your needs (fireballs,
    blasters shots, etc.).

    That class support the 2 types of representations: model and animations.
    The animation cases are slightly more evolved than the regular item.animation.
    It does use the item.animation but with more finesse as a projectile can travel in
    many directions. So it also keeps track of models and animation per travel
    direction.

    You probably want to subclass Projectile. It is totally ok to use it as it, but it
    is easier to create a subclass that contains all your Projectile information and let
    the game engine deal with orientation, range keeping, etc.
    Please see examples/07_projectiles.py for a good old fireball example.

    By default, Projectile travels in straight line in one direction. This behavior can
    be overwritten by setting a specific actuator (a projectile is a
    :class:`~pygamelib.board_items.Movable` so you can use my_projectile.actuator).

    The general way to use it is as follow:

    - Create a factory object with your static content (usually the static models,
      default direction and hit callback)
    - Add the direction related models and/or animation (keep in mind that animation
      takes precedence over static models)
    - deep copy that object when needed and add it to the projectiles stack of the game
      object.
    - use Game.actuate_projectiles(level) to let the Game engine do the heavy lifting.

    The Projectile constructor takes the following parameters:

    :param direction: A direction from the :ref:`constants-module` module
    :type direction: int
    :param range: The maximum range of the projectile in number of cells that can be
        crossed. When range is attained the hit_callback is called with a BoardItemVoid
        as a collision object.
    :type range: int
    :param step: the amount of cells a projectile can cross in one turn
    :type step: int
    :param model: the default model of the projectile.
    :type model: str
    :param movement_animation: the default animation of a projectile. If a projectile is
        sent in a direction that has no explicit and specific animation, then
        movement_animation is used if defined.
    :type movement_animation: :class:`~pygamelib.gfx.core.Animation`
    :param hit_animation: the animation used when the projectile collide with something.
    :type hit_animation: :class:`~pygamelib.gfx.core.Animation`
    :param hit_model: the model used when the projectile collide with something.
    :type hit_model: str
    :param hit_callback: A reference to a function that will be called upon collision.
        The hit_callback is receiving the object it collides with as first parameter.
    :type hit_callback: function
    :param is_aoe: Is this an 'area of effect' type of projectile? Meaning, is it doing
        something to everything around (mass heal, exploding rocket, fireball, etc.)?
        If yes, you must set that parameter to True and set the aoe_radius. If not, the
        Game object will only send the colliding object in front of the projectile.
    :type is_aoe: bool
    :param aoe_radius: the radius of the projectile area of effect. This will force the
        Game object to send a list of all objects in that radius.
    :type aoe_radius: int
    :param callback_parameters: A list of parameters to pass to hit_callback.
    :type callback_parameters: list
    :param movement_speed: The movement speed of the projectile
    :type movement_speed: int|float
    :param parent: The parent object (usually a Board object or some sort of BoardItem).

    .. important:: The effects of a Projectile are determined by the callback. No
        callback == no effect!

    Example::

        fireball = Projectile(
                                name="fireball",
                                model=Utils.red_bright(black_circle),
                                hit_model=Sprites.EXPLOSION,
                            )
        fireball.set_direction(constants.RIGHT)
        my_game.add_projectile(1, fireball,
                               my_game.player.pos[0], my_game.player.pos[1] + 1)


    """

    def __init__(
        self,
        name="projectile",
        direction=constants.RIGHT,
        step=1,
        range=5,
        model="\U00002301",
        movement_animation=None,
        hit_animation=None,
        hit_model=None,
        hit_callback=None,
        is_aoe=False,
        aoe_radius=0,
        parent=None,
        callback_parameters=[],
        movement_speed=0.15,
    ):
        if range % step != 0:
            raise base.PglException(
                "incorrect_range_step",
                "range must be a factor of step" " in Projectile",
            )
        Movable.__init__(
            self,
            model=model,
            step=step,
            name=name,
            parent=parent,
            movement_speed=movement_speed,
        )
        self.direction = direction
        self.range = range
        self.movement_animation = movement_animation
        self._directional_animations = {}
        self._directional_models = {}
        self.hit_animation = hit_animation
        self.hit_model = hit_model
        self.hit_callback = hit_callback
        self.callback_parameters = callback_parameters
        self.actuator = actuators.UnidirectionalActuator(direction=direction)
        self.is_aoe = is_aoe
        self.aoe_radius = aoe_radius
        self.parent = parent

    def add_directional_animation(self, direction, animation):
        """Add an animation for a specific direction.

        :param direction: A direction from the constants module.
        :type direction: int
        :param animation: The animation for the direction
        :type animation: :class:`~pygamelib.gfx.core.Animation`

        Example::

            fireball.add_directional_animation(constants.UP, constants.UP, animation)
        """
        if type(direction) is not int:
            raise base.PglInvalidTypeException(
                "Projectile.add_directional_animation "
                "requires an int from the constants module as"
                "direction."
            )
        if not isinstance(animation, core.Animation):
            raise base.PglInvalidTypeException(
                "Projectile.add_directional_animation "
                "requires a pygamelib.gfx.core.Animation as "
                "animation"
            )
        self._directional_animations[direction] = animation

    def directional_animation(self, direction):
        """Return the animation for a specific direction.

        :param direction: A direction from the constants module.
        :type direction: int
        :rtype: :class:`~pygamelib.gfx.core.Animation`

        Example::

            # No more animation for the UP direction
            fireball.directional_animation(constants.UP)
        """
        if type(direction) is not int:
            raise base.PglInvalidTypeException(
                "Projectile.add_directional_animation "
                "requires an int from the constants module as"
                "direction."
            )
        if direction in self._directional_animations:
            return self._directional_animations[direction]
        elif self.movement_animation is not None:
            return self.movement_animation
        else:
            return self.animation

    def remove_directional_animation(self, direction):
        """Remove an animation for a specific direction.

        :param direction: A direction from the constants module.
        :type direction: int

        Example::

            # No more animation for the UP direction
            fireball.remove_directional_animation(constants.UP)
        """
        if type(direction) is not int:
            raise base.PglInvalidTypeException(
                "Projectile.add_directional_animation "
                "requires an int from the constants module as"
                "direction."
            )
        del self._directional_animations[direction]

    def add_directional_model(self, direction, model):
        """Add an model for a specific direction.

        :param direction: A direction from the constants module.
        :type direction: int
        :param model: The model for the direction
        :type model: str

        Example::

            fireball.add_directional_animation(constants.UP, updward_animation)
        """
        if type(direction) is not int:
            raise base.PglInvalidTypeException(
                "Projectile.add_directional_model "
                "requires an int from the constants module as"
                "direction."
            )
        if type(model) is not str:
            raise base.PglInvalidTypeException(
                "Projectile.add_directional_model " "requires a string as model."
            )
        self._directional_models[direction] = model

    def directional_model(self, direction):
        """Return the model for a specific direction.

        :param direction: A direction from the constants module.
        :type direction: int
        :rtype: str

        Example::

            fireball.directional_model(constants.UP)
        """
        if type(direction) is not int:
            raise base.PglInvalidTypeException(
                "Projectile.add_directional_model "
                "requires an int from the constants module as"
                "direction."
            )
        if direction in self._directional_models:
            return self._directional_models[direction]
        else:
            return self.model

    def remove_directional_model(self, direction):
        """Remove the model for a specific direction.

        :param direction: A direction from the constants module.
        :type direction: int

        Example::

            fireball.directional_model(constants.UP)
        """
        if type(direction) is not int:
            raise base.PglInvalidTypeException(
                "Projectile.add_directional_model "
                "requires an int from the constants module as"
                "direction."
            )
        del self._directional_models[direction]

    def set_direction(self, direction):
        """Set the direction of a projectile

        This method will set a UnidirectionalActuator with the direction.
        It will also take care of updating the model and animation for the given
        direction if they are specified.

        :param direction: A direction from the constants module.
        :type direction: int

        Example::

            fireball.set_direction(constants.UP)
        """
        if type(direction) is not int:
            raise base.PglInvalidTypeException(
                "Projectile.set_direction "
                "requires an int from the constants module as"
                "direction."
            )
        self.model = self.directional_model(direction)
        self.animation = self.directional_animation(direction)
        self.direction = direction
        self.actuator = actuators.UnidirectionalActuator(direction=direction)

    def hit(self, objects):
        """A method that is called when the projectile hit something.

        That method is automatically called by the Game object when the Projectile
        collide with another object or is at the end of its range.

        Here are the call cases covered by the Game object:

         - range is reached without collision and projectile IS NOT an AoE type: hit()
           is called with a single BoardItemVoid in the objects list.
         - range is reached without collision and projectile IS an AoE type: hit()
           is called with the list of all objects within aoe_radius (including
           structures).
         - projectile collide with something and IS NOT an AoE type: hit() is called
           with the single colliding object in the objects list.
         - projectile collide with something and IS an AoE type: hit() is called with
           the list of all objects within aoe_radius (including structures).

        In turn, that method calls the hit_callback with the following parameters (in
        that order):

         1. the projectile object
         2. the list of colliding objects (that may contain only one object)
         3. the callback parameters (from the constructor callback_parameters)

        :param objects: A list of objects hit by or around the projectile.
        :type name: list

        Example::

            my_projectile.hit([npc1])

        """
        if self.hit_model is not None:
            self.model = self.hit_model
        if self.hit_animation is not None:
            self.animation = self.hit_animation
            self.animation.play_all()
        else:
            if self.animation is not None:
                self.animation.stop()
                self.animation = None
        self.actuator.stop()
        if self.hit_callback is not None:
            self.hit_callback(self, objects, self.callback_parameters)

    def has_inventory(self):
        """
        Projectile cannot have inventory by default.

        :return: False
        :rtype: Boolean
        """
        return False

    def overlappable(self):
        """
        Projectile are overlappable by default.

        :return: True
        :rtype: Boolean
        """
        return True

    def restorable(self):
        """
        We assume that by default, Projectiles are restorable.

        :return: True
        :rtype: bool
        """
        return True


class Immovable(BoardItem):
    """
    This class derive BoardItem and describe an object that cannot move or be
    moved (like a wall). Thus this class implements BoardItem.can_move().
    However it does not implement BoardItem.pickable() or
    BoardItem.overlappable()
    """

    def __init__(self, **kwargs):
        BoardItem.__init__(self, **kwargs)
        if "inventory_space" not in kwargs.keys():
            self._inventory_space = 0
        else:
            self._inventory_space = kwargs["inventory_space"]

    def can_move(self):
        """ Return the capability of moving of an item.

        Obviously an Immovable item is not capable of moving. So that method
        always returns False.

        :return: False
        :rtype: bool
        """
        return False

    def inventory_space(self):
        """Return the size of the Immovable Item for the
        :class:`~pygamelib.engine.Inventory`.

        :return: The size of the item.
        :rtype: int
        """
        return self._inventory_space

    def restorable(self):
        """
        This is a virtual method that must be implemented in deriving class.
        This method has to return True or False.
        This represent the capacity for an Immovable BoardItem to be restored
        by the board if the item is overlappable and has been overlapped by
        another Movable (:class:`~pygamelib.board_items.Movable`) item.
        """
        raise NotImplementedError()


class Actionable(Immovable):
    """
    This class derives :class:`~pygamelib.board_items.Immovable`. It adds the
    ability to an Immovable BoardItem to be triggered and execute some code.

    :param action: the reference to a function (Attention: no parentheses at
        the end of the function name).
    :type action: function
    :param action_parameters: the parameters to the action function.
    :type action_parameters: list
    :param perm: The permission that defines what types of items can actually
        activate the actionable. The permission has to be one of the
        permissions defined in :mod:`~pygamelib.constants`
    :type perm: :mod:`~pygamelib.constants`

    On top of these parameters Actionable accepts all parameters from
    :class:`~pygamelib.board_items.Immovable` and therefor from
    :class:`~pygamelib.board_items.BoardItem`.

    .. note:: The common way to use this class is to use
        GenericActionableStructure. Please refer to
        :class:`~pygamelib.board_items.GenericActionableStructure`
        for more details.
    """

    def __init__(self, **kwargs):
        if "action" not in kwargs.keys():
            self.action = None
        else:
            self.action = kwargs["action"]
        if "action_parameters" not in kwargs.keys():
            kwargs["action_parameters"] = []
        else:
            self.action_parameters = kwargs["action_parameters"]
        if "perm" not in kwargs.keys():
            self.perm = constants.PLAYER_AUTHORIZED
        else:
            self.perm = kwargs["perm"]
        Immovable.__init__(self, **kwargs)

    def activate(self):
        """
        This function is calling the action function with the
        action_parameters.

        Usually it's automatically called by :meth:`~pygamelib.engine.Board.move`
        when a Player or NPC (see :mod:`~pygamelib.board_items`)
        """
        if self.action is not None:
            self.action(self.action_parameters)


class Character:
    """A base class for a character (playable or not)

    :param agility: Represent the agility of the character
    :type agility: int
    :param attack_power: Represent the attack power of the character.
    :type attack_power: int
    :param defense_power: Represent the defense_power of the character
    :type defense_power: int
    :param hp: Represent the hp (Health Point) of the character
    :type hp: int
    :param intelligence: Represent the intelligence of the character
    :type intelligence: int
    :param max_hp: Represent the max_hp of the character
    :type max_hp: int
    :param max_mp: Represent the max_mp of the character
    :type max_mp: int
    :param mp: Represent the mp (Mana/Magic Point) of the character
    :type mp: int
    :param remaining_lives: Represent the remaining_lives of the character. For a NPC
        it is generally a good idea to set that to 1. Unless the NPC is a multi phased
        boss.
    :type remaining_lives: int
    :param strength: Represent the strength of the character
    :type strength: int

    These characteristics are here to be used by the game logic but very few of them are
    actually used by the Game (`pygamelib.engine`) engine.


    """

    def __init__(self, **kwargs):
        self.max_hp = None
        self.hp = None
        self.max_mp = None
        self.mp = None
        self.remaining_lives = None
        self.attack_power = None
        self.defense_power = None
        self.strength = None
        self.intelligence = None
        self.agility = None
        for a in [
            "max_hp",
            "hp",
            "max_mp",
            "mp",
            "remaining_lives",
            "attack_power",
            "defense_power",
            "strength",
            "intelligence",
            "agility",
        ]:
            if a in kwargs.keys():
                setattr(self, a, kwargs[a])


class Player(Movable, Character):
    """
    A class that represent a player controlled by a human.
    It accepts all the parameters from :class:`~pygamelib.board_items.Character` and is
    a :class:`~pygamelib.board_items.Movable`.

    This class sets a couple of variables to default values:

     * max_hp: 100
     * hp: 100
     * remaining_lives: 3
     * attack_power: 10
     * movement_speed: 0.1 (one movement every 0.1 second). Only useful if the game mode
         is set to MODE_RT.

    .. note:: If no inventory is passed as parameter a default one is created.
    """

    def __init__(self, **kwargs):
        if "max_hp" not in kwargs.keys():
            kwargs["max_hp"] = 100
        if "hp" not in kwargs.keys():
            kwargs["hp"] = 100
        if "remaining_lives" not in kwargs.keys():
            kwargs["remaining_lives"] = 3
        if "attack_power" not in kwargs.keys():
            kwargs["attack_power"] = 10
        if "movement_speed" not in kwargs.keys():
            kwargs["movement_speed"] = 0.1
        Movable.__init__(self, **kwargs)
        Character.__init__(self, **kwargs)
        if "inventory" in kwargs.keys():
            self.inventory = kwargs["inventory"]
        else:
            self.inventory = engine.Inventory(parent=self)

    def pickable(self):
        """This method returns False (a player is obviously not pickable).
        """
        return False

    def has_inventory(self):
        """This method returns True (a player has an inventory).
        """
        return True

    def overlappable(self):
        """This method returns false (a player cannot be overlapped).

        .. note:: If you wish your player to be overlappable, you need to inherit from
            that class and re-implement overlappable().
        """
        return False


class ComplexPlayer(Player, BoardComplexItem):
    """
    .. versionadded:: 1.2.0

    A complex player is nothing more than a :class:`~Player` mashed with a
    :class:`BoardComplexItem`.

    It supports all parameters of both with inheritance going first to Player and second
    to BoardComplexItem.

    The main interest is of course the multiple cell representation and the Sprites
    support.

    Example::

        player = ComplexPlayer(
                name='Mighty Wizard',
                sprite=sprite_collection['wizard_idle']
            )
    """

    def __init__(self, **kwargs):
        Player.__init__(self, **kwargs)
        BoardComplexItem.__init__(self, **kwargs)


class NPC(Movable, Character):
    """
    A class that represent a non playable character controlled by the computer.
    For the NPC to be successfully managed by the Game, you need to set an actuator.

    None of the parameters are mandatory, however it is advised to make good use of some
    of them (like type or name) for game design purpose.

    In addition to its own member variables, this class inherits all members from:
        * :class:`pygamelib.board_items.Character`
        * :class:`pygamelib.board_items.Movable`
        * :class:`pygamelib.board_items.BoardItem`

    This class sets a couple of variables to default values:

     * max_hp: 10
     * hp: 10
     * remaining_lives: 1
     * attack_power: 5
     * movement_speed: 0.25 (one movement every 0.25 second). Only useful if the game
         mode is set to MODE_RT.

    :param actuator: An actuator, it can be any class but it need to implement
        pygamelib.actuators.Actuator.
    :type actuator: pygamelib.actuators.Actuator

    Example::

        mynpc = NPC(name='Idiot McStupid', type='dumb_enemy')
        mynpc.step = 1
        mynpc.actuator = RandomActuator()
    """

    def __init__(self, **kwargs):
        if "max_hp" not in kwargs.keys():
            kwargs["max_hp"] = 10
        if "hp" not in kwargs.keys():
            kwargs["hp"] = 10
        if "remaining_lives" not in kwargs.keys():
            kwargs["remaining_lives"] = 1
        if "attack_power" not in kwargs.keys():
            kwargs["attack_power"] = 5
        if "movement_speed" not in kwargs.keys():
            kwargs["movement_speed"] = 0.25
        Movable.__init__(self, **kwargs)
        Character.__init__(self, **kwargs)
        if "actuator" not in kwargs.keys():
            self.actuator = None
        else:
            self.actuator = kwargs["actuator"]

        if "step" not in kwargs.keys():
            self.step = 1
        else:
            self.step = kwargs["step"]

    def pickable(self):
        """Define if the NPC is pickable.

        Obviously this method always return False.

        :return: False
        :rtype: Boolean

        Example::

            if mynpc.pickable():
                Utils.warn("Something is fishy, that NPC is pickable"
                    "but is not a Pokemon...")
        """
        return False

    def overlappable(self):
        """Define if the NPC is overlappable.

        Obviously this method also always return False.

        :return: False
        :rtype: Boolean

        Example::

            if mynpc.overlappable():
                Utils.warn("Something is fishy, that NPC is overlappable but"
                    "is not a Ghost...")
        """
        return False

    def has_inventory(self):
        """Define if the NPC has an inventory.

        This method returns false because the game engine doesn't manage NPC inventory
        yet but it could be in the future. It's a good habit to check the value returned
        by this function.

        :return: False
        :rtype: Boolean

        Example::

            if mynpc.has_inventory():
                print("Cool: we can pickpocket that NPC!")
            else:
                print("No pickpocketing XP for us today :(")
        """
        return False


class ComplexNPC(NPC, BoardComplexItem):
    """
    .. versionadded:: 1.2.0

    A complex NPC is nothing more than a :class:`~NPC` mashed with a
    :class:`BoardComplexItem`.

    It supports all parameters of both with inheritance going first to NPC and second
    to BoardComplexItem.

    The main interest is of course the multiple cell representation and the Sprites
    support.

    Example::

        player = ComplexNPC(
                name='Idiot McComplexStupid',
                sprite=npc_sprite_collection['troll_licking_stones']
            )
    """

    def __init__(self, **kwargs):
        NPC.__init__(self, **kwargs)
        BoardComplexItem.__init__(self, **kwargs)


class TextItem(BoardComplexItem):
    """
    .. versionadded:: 1.2.0

    The text item is a board item that can contains text. The text can then be
    manipulated and placed on a :class:`~pygamelib.engine.Board`.

    It is overall a :class:`BoardComplexItem` (so it takes all the parameters of that
    class). The big difference is that the first parameter is the text you want to
    display.

    The text parameter can be either a regular string or a :class:`~pygamelib.base.Text`
    object (in case you want formatting and colors).

    :param text: The text you want to display.
    :type text: str | :class:`~pygamelib.base.Text`

    Example::

        city_name = TextItem('Super City')
        fancy_city_name = TextItem(text=base.Text('Super City', base.Fore.GREEN,
            base.Back.BLACK,
            base.Style.BRIGHT
        ))
        my_board.place_item(city_name, 0, 0)
        my_board.place_item(fancy_city_name, 1, 0)
    """

    def __init__(self, text=None, **kwargs):
        BoardComplexItem.__init__(self, **kwargs)
        if text is not None and not isinstance(text, base.Text) and type(text) is str:
            self._text = base.Text(text)
        elif text is not None and not isinstance(text, base.Text):
            raise base.PglInvalidTypeException(
                f"TextItem: text parameter need to be either a str or a "
                f"pygamelib.base.Text object. Type {type(text)} is neither of these."
            )
        else:
            self._text = text
        self.sprite = core.Sprite.from_text(self._text)
        self.update_sprite()

    def __repr__(self):  # pragma: no cover
        return self._text.__repr__()

    @property
    def text(self):
        """The text within the item.

        TextItem.text can be set to either a string or a :class:`~pygamelib.base.Text`
        object.

        It will always return a :class:`~pygamelib.base.Text` object.

        Internally it translate the text to a :class:`~pygamelib.gfx.core.Sprite` to
        display it correctly on a :class:`~pygamelib.engine.Board`. If print()-ed it
        will do so like the :class:`~pygamelib.base.Text` object.
        """
        return self._text

    @text.setter
    def text(self, value):
        if type(value) is str:
            self._text = base.Text(value)
        elif isinstance(value, base.Text):
            self._text = value
        else:
            raise base.PglInvalidTypeException(
                "TextItem.text must be either a str or a pygamelib.base.Text object."
            )
        self.sprite = core.Sprite.from_text(self._text)
        self.update_sprite()


class Wall(Immovable):
    """
    A Wall is a specialized :class:`~pygamelib.board_items.Immovable` object that as
    unmodifiable characteristics:

    * It is not pickable (and cannot be).
    * It is not overlappable (and cannot be).
    * It is not restorable (and cannot be).

    As such it's an object that cannot be moved, cannot be picked up or modified by
    Player or NPC and block their ways. It is therefor advised to create one per board
    and reuse it in many places.

    :param model: The representation of the Wall on the Board.
    :type model: str
    :param name: The name of the Wall.
    :type name: str
    :param size: The size of the Wall. This parameter will probably be deprecated as
        size is only used for pickable objects.
    :type size: int
    """

    def __init__(self, **kwargs):
        if "model" not in kwargs.keys():
            kwargs["model"] = "#"
        if "name" not in kwargs.keys():
            kwargs["name"] = "wall"
        if "size" not in kwargs.keys():
            kwargs["size"] = 1
        Immovable.__init__(self, **kwargs)

    def pickable(self):
        """ This represent the capacity for a :class:`~pygamelib.board_items.BoardItem` to
        be pick-up by player or NPC.

        :return: False
        :rtype: bool

        Example::

            if mywall.pickable():
                print('Whoaa this wall is really light... and small...')
            else:
                print('Really? Trying to pick-up a wall?')
        """
        return False

    def overlappable(self):
        """ This represent the capacity for a :class:`~pygamelib.board_items.BoardItem` to
        be overlapped by player or NPC.

        :return: False
        :rtype: bool
        """
        return False

    def restorable(self):
        """
        This represent the capacity for an :class:`~pygamelib.board_items.Immovable`
        :class:`~pygamelib.board_items.BoardItem`(in this case a Wall item) to be
        restored by the board if the item is overlappable and has been overlapped by
        another :class:`~pygamelib.board_items.Movable` item.
        A wall is not overlappable.

        :return: False
        :rtype: bool
        """
        return False


class ComplexWall(Wall, BoardComplexItem):
    """
    .. versionadded:: 1.2.0

    A complex wall is nothing more than a :class:`~Wall` mashed with a
    :class:`BoardComplexItem`.

    It supports all parameters of both with inheritance going first to Wall and second
    to BoardComplexItem.

    The main interest is of course the multiple cell representation and the Sprites
    support.

    Example::

        wall = ComplexWall(
                sprite=sprite_brick_wall
            )
    """

    def __init__(self, **kwargs):
        Wall.__init__(self, **kwargs)
        BoardComplexItem.__init__(self, **kwargs)


class GenericStructure(Immovable):
    """
    A GenericStructure is as the name suggest, a generic object to create all kind of
    structures.

    It can be tweaked with all the properties of
    :class:`~pygamelib.board_items.BoardItem`, :class:`~pygamelib.board_items.Immovable`
    and it can be made pickable, overlappable or restorable or any combination of these.

    If you need an action to be done when a Player and/or a NPC touch the structure
    please have a look at
    :class:`pygamelib.board_items.GenericActionableStructure`.

    :param pickable: Define if the structure can be picked-up by a Player or NPC.
    :type pickable: bool
    :param overlappable: Define if the structure can be overlapped by a Player or NPC.
    :type overlappable: bool
    :param restorable: Define if the structure can be restored by the Board after a
        Player or NPC passed through. For example, you want a door or an activator
        structure (see GenericActionableStructure for that) to remain on the board after
        it's been overlapped by a player. But you could also want to develop some kind
        of Space Invaders game were the protection block are overlappable but not
        restorable.
    :type restorable: bool

    On top of these, this object takes all parameters of
    :class:`~pygamelib.board_items.BoardItem` and
    :class:`~pygamelib.board_items.Immovable`

    .. important:: If you need a structure with a permission system please have a look
        at :class:`~pygamelib.board_items.GenericActionableStructure`. This class
        has a permission system for activation.

    """

    def __init__(self, **kwargs):
        if "model" not in kwargs.keys():
            kwargs["model"] = "#"
        if "name" not in kwargs.keys():
            kwargs["name"] = "structure"
        Immovable.__init__(self, **kwargs)
        if "value" not in kwargs.keys():
            self.value = 0
        else:
            self.value = kwargs["value"]
        if "pickable" in kwargs.keys():
            self.__is_pickable = kwargs["pickable"]
        else:
            self.__is_pickable = False
        if "overlappable" in kwargs.keys():
            self.__is_overlappable = kwargs["overlappable"]
        else:
            self.__is_overlappable = False

        if "restorable" in kwargs.keys():
            self.__is_restorable = kwargs["restorable"]
        else:
            self.__is_restorable = False

    def pickable(self):
        """This represent the capacity for a BoardItem to be picked-up by player or NPC.

        To set this value please use :meth:`~.set_pickable`

        :return: True or False
        :rtype: bool

        .. seealso:: :meth:`~.set_pickable`
        """
        return self.__is_pickable

    def set_pickable(self, val):
        """Make the structure pickable or not.

        :param val: True or False depending on the pickability of the structure.
        :type val: bool

        Example::

            myneatstructure.set_pickable(True)
        """
        if type(val) is bool:
            self.__is_pickable = val

    def overlappable(self):
        """ This represent the capacity for a :class:`~pygamelib.board_items.BoardItem` to
        be overlapped by player or NPC.

        To set this value please use :meth:`~.set_overlappable`

        :return: False
        :rtype: bool

        .. seealso:: :meth:`~.set_overlappable`
        """
        return self.__is_overlappable

    def set_overlappable(self, val):
        """Make the structure overlappable or not.

        :param val: True or False depending on the fact that the structure can be
            overlapped (i.e that a Player or NPC can step on it) or not.
        :type val: bool

        Example::

            myneatstructure.set_overlappable(True)
        """
        if type(val) is bool:
            self.__is_overlappable = val

    def restorable(self):
        """
        This represent the capacity for an :class:`~pygamelib.board_items.Immovable`
        :class:`~pygamelib.board_items.BoardItem` (in this case a GenericStructure item)
        to be restored by the board if the item is overlappable and has been overlapped
        by another :class:`~pygamelib.board_items.Movable` item.

        The value of this property is set with :meth:`~.set_restorable`

        :return: False
        :rtype: bool

        .. seealso:: :meth:`~.set_restorable`
        """
        return self.__is_restorable

    def set_restorable(self, val):
        """Make the structure restorable or not.

        :param val: True or False depending on the restorability of the structure.
        :type val: bool

        Example::

            myneatstructure.set_restorable(True)
        """
        if type(val) is bool:
            self.__is_restorable = val
        else:
            raise base.PglInvalidTypeException(
                "set_restorable(bool) takes a boolean as paramater."
            )


class GenericActionableStructure(GenericStructure, Actionable):
    """
    A GenericActionableStructure is the combination of a
    :class:`~pygamelib.board_items.GenericStructure` and an
    :class:`~pygamelib.board_items.Actionable`.
    It is only a helper combination.

    Please see the documentation for
    :class:`~pygamelib.board_items.GenericStructure` and
    :class:`~pygamelib.board_items.Actionable` for more information.
    """

    def __init__(self, **kwargs):
        GenericStructure.__init__(self, **kwargs)
        Actionable.__init__(self, **kwargs)


class Treasure(Immovable):
    """
    A Treasure is an :class:`~pygamelib.board_items.Immovable` that is pickable and
    with a non zero value. It is an helper class that allows to focus on game design and
    mechanics instead of small building blocks.

    :param model: The model that will represent the treasure on the map
    :type model: str
    :param value: The value of the treasure, it is usually used to calculate the score.
    :type value: int
    :param inventory_space: The space occupied by the treasure. It is used by
        :class:`~pygamelib.engine.Inventory` as a measure of space. If the treasure's
        size exceed the Inventory size (or the cumulated size of all items + the
        treasure exceed the inventory max_size()) the
        :class:`~pygamelib.base.Inventory` will refuse to add the treasure.
    :type inventory_space: int

    .. note:: All the options from :class:`~pygamelib.board_items.Immovable` are also
        available to this constructor.

    Example::

        money_bag = Treasure(model=Sprites.MONEY_BAG,value=100,inventory_space=2)
        print(f"This is a money bag {money_bag}")
        player.inventory.add_item(money_bag)
        print(f"The inventory value is {player.inventory.value()} and is at
            {player.inventory.size()}/{player.inventory.max_size}")
    """

    def __init__(self, **kwargs):
        if "model" not in kwargs.keys():
            kwargs["model"] = "¤"
        Immovable.__init__(self, **kwargs)
        if "value" not in kwargs.keys():
            self.value = 10
        else:
            self.value = kwargs["value"]
        if "inventory_space" not in kwargs.keys():
            self._inventory_space = 1
        else:
            self._inventory_space = kwargs["inventory_space"]

    def pickable(self):
        """ This represent the capacity for a Treasure to be picked-up by player or NPC.

        A treasure is obviously pickable by the player and potentially NPCs.
        :class:`~pygamelib.engine.Board` puts the Treasure in the
        :class:`~pygamelib.engine.Inventory` if the picker implements has_inventory()

        :return: True
        :rtype: bool
        """
        return True

    def overlappable(self):
        """ This represent the capacity for a Treasure to be overlapped by player or NPC.

        A treasure is not overlappable.

        :return: False
        :rtype: bool
        """
        return False

    def restorable(self):
        """ This represent the capacity for a Treasure to be restored after being overlapped.

        A treasure is not overlappable, therefor is not restorable.

        :return: False
        :rtype: bool
        """
        return False


class ComplexTreasure(Treasure, BoardComplexItem):
    """
    .. versionadded:: 1.2.0

    A complex treasure is nothing more than a :class:`~Treasure` mashed with a
    :class:`BoardComplexItem`.

    It supports all parameters of both with inheritance going first to Treasure and
    second to BoardComplexItem.

    The main interest is of course the multiple cell representation and the Sprites
    support.

    Example::

        chest = ComplexTreasure(
                sprite=sprite_chest
            )
    """

    def __init__(self, **kwargs):
        Treasure.__init__(self, **kwargs)
        BoardComplexItem.__init__(self, **kwargs)


class Door(GenericStructure):
    """
    A Door is a :class:`~pygamelib.board_items.GenericStructure` that is not
    pickable, overlappable and restorable. It has a value of 0 and a size of 1 by
    default. It is an helper class that allows to focus on game design and mechanics
    instead of small building blocks.

    :param model: The model that will represent the door on the map
    :type model: str
    :param value: The value of the door, it is useless in that case. The default value
        is 0.
    :type value: int
    :param inventory_space: The size of the door in the inventory. Unless you make the
        door pickable (I have no idea why you would do that...), this parameter is not
        used.
    :type inventory_space: int
    :param type: The type of the door. It is often used as a type identifier for your
        game main loop. For example: unlocked_door or locked_door.
    :type type: str
    :param pickable: Is this door pickable by the player? Default value is False.
    :type pickable: Boolean
    :param overlappable: Is this door overlappable by the player? Default value is True.
    :type overlappable: Boolean
    :param restorable: Is this door restorable after being overlapped? Default value is
        True.
    :type restorable: Boolean

    .. note:: All the options from
       :class:`~pygamelib.board_items.GenericStructure` are also available to
       this constructor.

    Example::

        door1 = Door(model=Sprites.DOOR,type='locked_door')
    """

    def __init__(self, **kwargs):
        if "model" not in kwargs.keys():
            kwargs["model"] = "]"
        Immovable.__init__(self, **kwargs)
        if "value" not in kwargs.keys():
            self.value = 0
        else:
            self.value = kwargs["value"]
        if "inventory_space" not in kwargs.keys():
            self._inventory_space = 1
        else:
            self._inventory_space = kwargs["inventory_space"]
        if "name" not in kwargs.keys():
            self.name = "Door"
        else:
            self.name = kwargs["name"]
        if "type" not in kwargs.keys():
            self.type = "door"
        else:
            self.type = kwargs["type"]
        if "pickable" not in kwargs.keys():
            self.set_pickable(False)
        else:
            self.set_pickable(kwargs["pickable"])
        if "overlappable" not in kwargs.keys():
            self.set_overlappable(True)
        else:
            self.set_overlappable(kwargs["overlappable"])
        if "restorable" not in kwargs.keys():
            self.set_restorable(True)
        else:
            self.set_restorable(kwargs["restorable"])


class ComplexDoor(Door, BoardComplexItem):
    """
    .. versionadded:: 1.2.0

    A complex door is nothing more than a :class:`~Door` mashed with a
    :class:`BoardComplexItem`.

    It supports all parameters of both with inheritance going first to Door and second
    to BoardComplexItem.

    The main interest is of course the multiple cell representation and the Sprites
    support.

    Example::

        castle_door = ComplexDoor(
                sprite=sprite_castle_door
            )
    """

    def __init__(self, **kwargs):
        Door.__init__(self, **kwargs)
        BoardComplexItem.__init__(self, **kwargs)


class GenericStructureComplexComponent(GenericStructure, BoardItemComplexComponent):
    """
    A ComplexComponent specifically for generic structures.
    """

    def __init__(self, **kwargs):
        GenericStructure.__init__(self, **kwargs)
        BoardItemComplexComponent.__init__(self, **kwargs)


class Tile(GenericStructure, BoardComplexItem):
    """
    .. versionadded:: 1.2.0

    A Tile is a standard :class:`BoardComplexItem` configured by default to:

     * be overlappable
     * be not pickable
     * be immovable.

    Aside from the movable attributes (it inherit from GenericStructure so it's an
    Immovable object), everything else is configurable.

    It is particularly useful to display a :class:`~pygamelib.gfx.core.Sprite` on the
    background or to create terrain.

    :param overlappable: Defines if the Tile can be overlapped.
    :type overlappable: bool
    :param restorable: Defines is the Tile should be restored after being overlapped.
    :type restorable: bool
    :param pickable: Defines if the Tile can be picked up by the Player or NPC.
    :type pickable: bool

    Please see :class:`BoardComplexItem` for additional parameters.

    Example::

        grass_sprite = Sprite.load_from_ansi_file('textures/grass.ans')
        for pos in grass_positions:
            outdoor_level.place_item( Tile(sprite=grass_sprite), pos[0], pos[1] )
    """

    def __init__(self, **kwargs):
        kwargs["parent"] = self
        kwargs["base_item_type"] = GenericStructureComplexComponent
        if "overlappable" not in kwargs:
            kwargs["overlappable"] = True
        if "restorable" not in kwargs:
            kwargs["restorable"] = True
        if "pickable" not in kwargs:
            kwargs["pickable"] = False
        GenericStructure.__init__(self, **kwargs)
        # kwargs["base_item_type"] = Door
        BoardComplexItem.__init__(self, **kwargs)

    def can_move(self):
        """A Tile cannot move.

        :returns: False
        :rtype: bool
        """
        return False



"""This module contains the base classes for simple and advanced actuators.
These classes are the base contract for actuators.
If you wish to create your own one, you need to inheritate from one of these base class.

This module contains the simple actuators classes.
Simple actuators are movement related one. They allow for predetermined movements
patterns.

.. autosummary::
   :toctree: .

    Actuator
    Behavioral
    RandomActuator
    PathActuator
    PatrolActuator
    UnidirectionalActuator
    PathFinder

"""
from pygamelib import board_items
from pygamelib import base
from pygamelib import constants
import random
import collections


class Actuator:
    """
    Actuator is the base class for all Actuators. It is mainly a contract class with
    some utility methods.

    By default, all actuators are considered movement actuators. So the base class only
    require next_move() to be implemented.

    :param parent: the item parent.

    """

    def __init__(self, parent):
        """
        The constructor take only one (positional) parameter: the parent object.

        .. important:: The default state of ALL actuators is RUNNING. If you want your
            actuator to be in a different state (PAUSED for example), you have to do it
            yourself.
        """
        self.type = None
        self.state = constants.RUNNING
        self.parent = parent

    def start(self):
        """Set the actuator state to RUNNING.

        If the actuator state is not RUNNING, actuators' next_move() function
        (and all derivatives) should not return anything.

        Example::

            mygame.start()
        """
        self.state = constants.RUNNING

    def pause(self):
        """Set the actuator state to PAUSED.

        Example::

            mygame.pause()
        """
        self.state = constants.PAUSED

    def stop(self):
        """Set the actuator state to STOPPED.

        Example::

            mygame.stop()
        """
        self.state = constants.STOPPED

    def next_move(self):
        """
        That method needs to be implemented by all actuators or a NotImplementedError
        exception will be raised.

        :raises: NotImplementedError
        """
        raise NotImplementedError()


class Behavioral(Actuator):
    """
    The behavioral actuator is inheriting from Actuator and is adding a next_action()
    method.
    The actual actions are left to the actuator that implements Behavioral.

    :param parent: the item parent.

    """

    def __init__(self, parent):
        """
        The constructor simply construct an Actuator. It takes on positional paraneter:
        the parent object.
        """
        Actuator.__init__(self, parent)

    def next_action(self):
        """
        That method needs to be implemented by all behavioral actuators or a
        NotImplementedError exception will be raised.

        :raises: NotImplementedError
        """
        raise NotImplementedError()


class RandomActuator(Actuator):
    """A class that implements a random choice of movement.

    The random actuator is a subclass of
    :class:`~pygamelib.actuators.Actuator`.
    It is simply implementing a random choice in a predefined move set.

    :param moveset: A list of movements.
    :type moveset: list
    :param parent: The parent object to actuate.
    :type parent: pygamelib.board_items.BoardItem
    """

    def __init__(self, moveset=None, parent=None):
        if moveset is None:
            moveset = []
        super().__init__(parent)
        self.moveset = moveset

    def next_move(self):
        """Return a randomly selected movement

        The movement is randomly selected from moveset if state is RUNNING,
        otherwise it should return None.

        :return: The next movement
        :rtype: int | None

        Example::

            randomactuator.next_move()
        """
        if self.state == constants.RUNNING:
            return random.choice(self.moveset)


class PathActuator(Actuator):
    """
    The path actuator is a subclass of
    :class:`~pygamelib.actuators.Actuator`.
    The move inside the function next_move
    depends on path and index. If the state is not running it returns None
    otherwise it increments the index & then, further compares the index
    with length of the path. If they both are same then, index is set to
    value zero and the move is returned back.

    :param path: A list of paths.
    :type path: list
    :param parent: The parent object to actuate.
    :type parent: pygamelib.board_items.BoardItem
    """

    def __init__(self, path=None, parent=None):
        if path is None:
            path = []  # pragma: no cover
        super().__init__(parent)
        self.path = path
        self.index = 0

    def next_move(self):
        """Return the movement based on current index

        The movement is selected from path if state is RUNNING, otherwise
        it should return None. When state is RUNNING, the movement is selected
        before incrementing the index by 1. When the index equal the length of
        path, the index should return back to 0.

        :return: The next movement
        :rtype: int | None

        Example::

            pathactuator.next_move()
        """
        if self.state == constants.RUNNING:
            move = self.path[self.index]
            self.index += 1
            if self.index == len(self.path):
                self.index = 0
            return move

    def set_path(self, path):
        """Defines a new path

        This will also reset the index back to 0.

        :param path: A list of movements.
        :type path: list

        Example::

            pathactuator.set_path([constants.UP,constants.DOWN,constants.LEFT,constants.RIGHT])
        """
        self.path = path
        self.index = 0


class PatrolActuator(PathActuator):
    """
    The patrol actuator is a subclass of
    :class:`~pygamelib.actuators.PathActuator`.  The move inside the function
    next_move depends on path and index and the mode. Once it reaches the end
    of the move list it will start cycling back to the beggining of the list.
    Once it reaches the beggining it will start moving forwards
    If the state is not running it returns None otherwise it increments the
    index & then, further compares the index with length of the path.
    If they both are same then, index is set to value zero and the move is
    returned back.

    :param path: A list of directions.
    :type path: list
    """

    def next_move(self):
        """Return the movement based on current index

        The movement is selected from path if state is RUNNING, otherwise it
        should return None. When state is RUNNING, the movement is selected
        before incrementing the index by 1. When the index equals the length
        of path, the index should return back to 0 and the path list should be
        reversed before the next call.

        :return: The next movement
        :rtype: int | None

        Example::

            patrolactuator.next_move()
        """
        if self.state == constants.RUNNING:
            move = self.path[self.index]
            self.index += 1
            if self.index == len(self.path):
                self.index = 0
                self.path.reverse()
                for i in range(0, len(self.path)):
                    if self.path[i] == constants.UP:
                        self.path[i] = constants.DOWN
                    elif self.path[i] == constants.DOWN:
                        self.path[i] = constants.UP
                    elif self.path[i] == constants.LEFT:
                        self.path[i] = constants.RIGHT
                    elif self.path[i] == constants.RIGHT:
                        self.path[i] = constants.LEFT
                    elif self.path[i] == constants.DLDOWN:
                        self.path[i] = constants.DRUP
                    elif self.path[i] == constants.DLUP:
                        self.path[i] = constants.DRDOWN
                    elif self.path[i] == constants.DRDOWN:
                        self.path[i] = constants.DLUP
                    elif self.path[i] == constants.DRUP:
                        self.path[i] = constants.DLDOWN
            return move


class UnidirectionalActuator(Actuator):
    """A class that implements a single movement.

    The unidirectional actuator is a subclass of
    :class:`~pygamelib.actuators.Actuator`.
    It is simply implementing a mono directional movement. It is primarily target at
    projectiles.

    :param direction: A single direction from the Constants module.
    :type direction: int
    :param parent: The parent object to actuate.
    :type parent: pygamelib.board_items.BoardItem
    """

    def __init__(self, direction=constants.RIGHT, parent=None):
        if direction is None:
            direction = constants.RIGHT
        super().__init__(parent)
        self.direction = direction

    def next_move(self):
        """Return the direction.

        The movement is always direction if state is RUNNING,
        otherwise it returns None.

        :return: The next movement
        :rtype: int | None

        Example::

            unidirectional_actuator.next_move()
        """
        if self.state == constants.RUNNING:
            return self.direction


class PathFinder(Behavioral):
    """
    .. important:: This module assume a one step movement.
        If you need more than one step, you will need to sub-class
        this module and re-implement next_waypoint().

    This actuator is a bit different than the simple actuators
    (:mod:`~pygamelib.actuators.SimpleActuators`) as it requires
    the knowledge of both the game object and the actuated object.

    The constructor takes the following parameters:

    :param game: A reference to the instanciated game engine.
    :type game: pygamelib.engine.Game
    :param actuated_object: The object to actuate. Deprecated in favor of parent.
        Only kept for backward compatibility.
    :type actuated_object: pygamelib.board_items.BoardItem
    :param parent: The parent object to actuate.
    :type parent: pygamelib.board_items.BoardItem
    :param circle_waypoints: If True the next_waypoint()
        method is going to circle between the waypoints
        (when the last is visited, go back to the first)
    :type circle_waypoints: bool

    """

    def __init__(
        self, game=None, actuated_object=None, circle_waypoints=True, parent=None
    ):
        if actuated_object is not None and parent is None:
            self.actuated_object = actuated_object
            self.parent = actuated_object
        elif parent is not None:
            self.actuated_object = parent
            self.parent = parent
        super().__init__(self.parent)
        self.destination = (None, None)
        self.game = game
        self._current_path = []
        self.waypoints = []
        self._waypoint_index = 0
        self.circle_waypoints = circle_waypoints

    def set_destination(self, row=0, column=0):
        """Set the targeted destination.

        :param row: "row" coordinate on the board grid
        :type row: int
        :param column: "column" coordinate on the board grid
        :type column: int
        :raises PglInvalidTypeException: if row or column are not int.

        Example::

            mykillernpc.actuator.set_destination(
                mygame.player.pos[0], mygame.player.pos[1]
            )
        """
        if type(row) is not int or type(column) is not int:
            raise base.PglInvalidTypeException(
                "In Actuator.PathFinder.set_destination(x,y) both x and y must be "
                "integer."
            )
        self.destination = (row, column)

    def find_path(self):
        """Find a path to the destination.

        Destination (PathFinder.destination) has to be set beforehand.
        This method implements a Breadth First Search algorithm
        (`Wikipedia <https://en.wikipedia.org/wiki/Breadth-first_search>`_)
        to find the shortest path to destination.

        Example::

            mykillernpc.actuator = PathFinder(
                    game=mygame, actuated_object=mykillernpc
                )
            mykillernpc.actuator.set_destination(
                    mygame.player.pos[0], mygame.player.pos[1]
                )
            mykillernpc.actuator.find_path()

        .. warning:: PathFinder.destination is a tuple!
            Please use PathFinder.set_destination(x,y) to avoid problems.

        """
        if self.actuated_object is None:
            raise base.PglException(
                "actuated_object is not defined",
                "PathFinder.actuated_object has to be defined.",
            )
        if not isinstance(self.actuated_object, board_items.Movable):
            raise base.PglException(
                "actuated_object not a Movable object",
                "PathFinder.actuated_object has to be an instance \
                of a Movable object.",
            )
        if self.destination is None:
            raise base.PglException(
                "destination is not defined",
                "PathFinder.destination has to be defined.",
            )

        queue = collections.deque(
            [[(self.actuated_object.pos[0], self.actuated_object.pos[1])]]
        )
        seen = set([(self.actuated_object.pos[0], self.actuated_object.pos[1])])
        while queue:
            path = queue.popleft()
            x, y = path[-1]
            if (x, y) == self.destination:
                self._current_path = path
                # We return only a copy of the path as we need to keep the
                # real one untouched for our own needs.
                return path.copy()
            # r = row c = column
            for r, c in ((x + 1, y), (x - 1, y), (x, y + 1), (x, y - 1)):
                if (
                    0 <= c < self.game.current_board().size[0]
                    and 0 <= r < self.game.current_board().size[1]
                    and self.game.current_board().item(r, c).overlappable()
                    and (r, c) not in seen
                ):
                    queue.append(path + [(r, c)])
                    seen.add((r, c))
        return []

    def current_path(self):
        """This method simply return a copy of the current path of the actuator.

        The current path is to be understood as: the list of positions still
        remaining. All positions that have already been gone through are
        removed from the stack.

        .. important:: A copy of the path is returned for every call to that
            function so be wary of the performances impact.

        Example::

            mykillernpc.actuator = PathFinder(
                                    game=mygame,
                                    actuated_object=mykillernpc
                                )
            mykillernpc.actuator.set_destination(
                                    mygame.player.pos[0],
                                    mygame.player.pos[1]
                                )
            mykillernpc.actuator.find_path()
            for i in mykillernpc.actuator.current_path():
                print(i)

        """
        return self._current_path.copy()

    # That function is actually covered in tests but some cases are too edgy to test.
    def next_move(self):  # pragma: no cover
        """This method return the next move calculated by this actuator.

        In the case of this PathFinder actuator, next move does the following:

         - If the destination is not set return NO_DIR \
            (see :py:mod:`~pygamelib.constants`) \
         - If the destination is set, but the path is empty and actuated \
            object's position is different from destination: \
            call :meth:`find_path()`
         - Look at the current waypoint, if the actuated object is not at \
            that position return a direction from the \
            :mod:`~pygamelib.constants` module. The direction is calculated \
            from the difference betwen actuated object's position and \
            waypoint's position.
         - If the actuated object is at the waypoint position, then call \
            next_waypoint(), set the destination and return a direction. \
            In this case, also call :meth:`find_path()`.
         - In any case, if there is no more waypoints in the path this method \
            returns NO_DIR (see :py:mod:`~pygamelib.constants`)

        Example::

            seeker = NPC(model=Sprites.SKULL)
            seeker.actuator = PathFinder(game=mygame,actuated_object=seeker)
            while True:
                seeker.actuator.set_destination(mygame.player.pos[0],mygame.player.pos[1])
                # next_move() will call find_path() for us.
                next_move = seeker.actuator.next_move()
                if next_move == constants.NO_DIR:
                    seeker.actuator.set_destination(mygame.player.pos[0],mygame.player.pos[1])
                else:
                    mygame.current_board().move(seeker,next_move,1)
        """
        # If one of destination coordinate is None, return NO_DIR
        if self.destination[0] is None or self.destination[1] is None:
            return constants.NO_DIR

        # If path is empty and actuated_object is not at destination,
        # try to find a path to destination
        if len(self._current_path) == 0 and (
            self.actuated_object.pos[0] != self.destination[0]
            or self.actuated_object.pos[1] != self.destination[1]
        ):
            self.find_path()

        # If path is still empty return NO_DIR (destination is unreachable or
        # the current waypoint is reached)
        if len(self._current_path) == 0:
            # First we check if we already are at current waypoint
            (cwr, cwc) = self.current_waypoint()
            if (
                self.actuated_object.pos[0] == cwr
                and self.actuated_object.pos[1] == cwc
            ):
                # If so, we get the next waypoint
                (r, c) = self.next_waypoint()
                # If there are no more waypoints, then we return NO_DIR
                if r is None or c is None:
                    return constants.NO_DIR
                else:
                    # Else we set the new destination and calculate the path
                    self.set_destination(r, c)
                    self.find_path()
            else:
                return constants.NO_DIR

        if len(self._current_path) == 0:
            return constants.NO_DIR

        # Get the next position from the path
        next_position = self._current_path.pop(0)
        # If actuated object is already there check if the path is empty, if so
        # return NO_DIR, else get the next position from the path.
        if (
            next_position[0] == self.actuated_object.pos[0]
            and next_position[1] == self.actuated_object.pos[1]
        ):
            if len(self._current_path) == 0:
                return constants.NO_DIR
            next_position = self._current_path.pop(0)

        # print(f'Next position is: {next_position} and object position is
        # {self.actuated_object.pos} _current_path length
        # is {len(self._current_path)}')
        dr = self.actuated_object.pos[0] - next_position[0]
        dc = self.actuated_object.pos[1] - next_position[1]

        # Look at the difference between the current position and next position
        # and return the correct direction.
        # If the coordinates are impossible to resolve return NO_DIR
        # NOTE: We could use a comparison with 0
        #   (dr < 0 instead of dr == -1 for example) but the sides effects are
        #   numerous in case the user use a step higher than 1 in Board.move().
        #   The actuated object could end up going into a wall or out of bound.
        if dr == -1 and dc == 0:
            return constants.DOWN
        elif dr == 1 and dc == 0:
            return constants.UP
        elif dr == 0 and dc == 1:
            return constants.LEFT
        elif dr == 0 and dc == -1:
            return constants.RIGHT
        elif dr == -1 and dc == -1:
            return constants.DRDOWN
        elif dr == 1 and dc == -1:
            return constants.DRUP
        elif dr == -1 and dc == 1:
            return constants.DLDOWN
        elif dr == 1 and dc == 1:
            return constants.DLUP
        elif dr > 1 or dr < -1 or dc > 1 or dc < -1:
            # If we are here it means that something is blocking the movement
            self.find_path()
            return self.next_move()
        else:
            return constants.NO_DIR

    def add_waypoint(self, row, column):
        """Add a waypoint to the list of waypoints.

        Waypoints are used one after the other on a FIFO basis
        (First In, First Out).

        If not destination (i.e destination == (None, None)) have been set yet, that
        method sets it.

        :param row: The "row" part of the waypoint's coordinate.
        :type row: int
        :param column: The "column" part of the waypoint's coordinate.
        :type row: int
        :raise PglInvalidTypeException: If any of the parameters is not an int.

        Example::

            pf = PathFinder(game=mygame, actuated_object=npc1)
            pf.add_waypoint(3,5)
            pf.add_waypoint(12,15)

        """
        if type(row) is not int:
            raise base.PglInvalidTypeException('"row" is not an integer. It must be.')
        if type(column) is not int:
            raise base.PglInvalidTypeException(
                '"column" is not an integer. It must be.'
            )
        self.waypoints.append((row, column))
        if self.destination == (None, None):
            self.destination = (row, column)

    def clear_waypoints(self):
        """Empty the waypoints stack.

        Example::

            pf.clear_waypoints()
        """
        self.waypoints.clear()
        self._waypoint_index = -1

    def current_waypoint(self):
        """Return the currently active waypoint.

        If no waypoint have been added, this function return None.

        :return: Either a None tuple or the current waypoint.
        :rtype: A None tuple or a tuple of integer.

        Example::

            (row,column) = pf.current_waypoint()
            pf.set_destination(row,column)

        """
        if len(self.waypoints) == 0:
            return (None, None)
        return self.waypoints[self._waypoint_index]

    def next_waypoint(self):
        """Return the next active waypoint.

        If no waypoint have been added, this function return None.
        If there is no more waypoint in the stack:

        - if PathFinder.circle_waypoints is True this function reset the \
        waypoints stack and return the first one.
        - else, return None.

        :return: Either a None tuple or the next waypoint.
        :rtype: A None tuple or a tuple of integer.

        Example::

            pf.circle_waypoints = True
            (row,column) = pf.next_waypoint()
            pf.set_destination(row,column)

        """
        if len(self.waypoints) == 0:
            return (None, None)
        self._waypoint_index += 1
        if self._waypoint_index >= len(self.waypoints):
            if self.circle_waypoints:
                self._waypoint_index = 0
            else:
                return (None, None)
        return self.waypoints[self._waypoint_index]

    def remove_waypoint(self, row, column):
        """ Remove a waypoint from the stack.

        This method removes the first occurrence of a waypoint in the stack.

        If the waypoint cannot be found, it raises a ValueError exception.
        If the row and column parameters are not int, an
        PglInvalidTypeException is raised.

        :param row: The "row" part of the waypoint's coordinate.
        :type row: int
        :param column: The "column" part of the waypoint's coordinate.
        :type row: int
        :raise PglInvalidTypeException: If any of the parameters is not an int.
        :raise ValueError: If the waypoint is not found in the stack.

        Example::

            method()
        """
        if type(row) is not int:
            raise base.PglInvalidTypeException('"row" is not an integer. It must be.')
        if type(column) is not int:
            raise base.PglInvalidTypeException(
                '"column" is not an integer. It must be.'
            )
        try:
            idx = self.waypoints.index((row, column))
            self.waypoints.pop(idx)
        except ValueError:
            raise base.PglException(
                "invalid_waypoint",
                f"Waypoint ({row},{column}) does not exist in the waypoints stack.",
            )



#!/usr/bin/env python3

import pygamelib.engine as engine
import pygamelib.board_items as board_items
import pygamelib.assets.graphics as graphics
import pygamelib.constants as constants
import time
import sys

board_to_load = "hac-maps/test-board.json"

if len(sys.argv) > 1:
    board_to_load = sys.argv[1]

max_iter = 10
if len(sys.argv) > 2:
    max_iter = int(sys.argv[2])


g = engine.Game()

b = g.load_board(board_to_load, 1)

if b.width >= g.screen.width or b.height >= g.screen.height:
    g.enable_partial_display = True
    g.partial_display_viewport = [
        int((g.screen.height - 2) / 2),
        int((g.screen.width - 2) / 4),
    ]

g.player = board_items.Player(model=graphics.Models.FLYING_SAUCER)
g.player.inventory.max_size = 99999
g.change_level(1)


idx = 0
key = None

while idx < max_iter or max_iter == 0:
    if key == "w":
        g.move_player(constants.UP, 1)
    elif key == "s":
        g.move_player(constants.DOWN, 1)
    elif key == "a":
        g.move_player(constants.LEFT, 1)
    elif key == "d":
        g.move_player(constants.RIGHT, 1)
    elif key == "q":
        break
    g.clear_screen()
    g.actuate_npcs(1)
    g.display_board()

    if max_iter == 0:
        key = engine.Game.get_key()
    else:
        time.sleep(0.1)
        idx += 1



"""
Accessible constants are the following:

General purpose:
 * PYGAMELIB_VERSION

Directions:
 * NO_DIR : This one is used when no direction can be provided by an actuator
    (destination reached for a PathFinder for example)
 * UP
 * DOWN
 * LEFT
 * RIGHT
 * DRUP : Diagonal right up
 * DRDOWN : Diagonal right down
 * DLUP : Diagonal Left up
 * DLDOWN : Diagonal left down

Permissions:
 * PLAYER_AUTHORIZED
 * NPC_AUTHORIZED
 * ALL_PLAYABLE_AUTHORIZED (deprecated in 1.2.0 in favor of ALL_CHARACTERS_AUTHORIZED)
 * ALL_CHARACTERS_AUTHORIZED
 * ALL_MOVABLE_AUTHORIZED
 * NONE_AUTHORIZED

UI positions:
 * POS_TOP
 * POS_BOTTOM
 * ORIENTATION_HORIZONTAL
 * ORIENTATION_VERTICAL

Actions states (for Actuators for example):
 * RUNNING
 * PAUSED
 * STOPPED

Special constants:
 * NO_PLAYER: That constant is used to tell the Game object not to manage the player.
 * MODE_RT: Set the game object to Real Time mode. The game runs independently from the
    user input.
 * MODE_TBT: Set the game object to Turn By Turn mode. The game runs turn by turn and
    pause between each user input.
"""

# Main version
PYGAMELIB_VERSION = "1.2.3"

# Directions
NO_DIR = 10000000
UP = 10000001
DOWN = 10000010
LEFT = 10000011
RIGHT = 10000100
DRUP = 10000101
DRDOWN = 10000110
DLUP = 10000111
DLDOWN = 10001000

# Permissions
PLAYER_AUTHORIZED = 20000001
NPC_AUTHORIZED = 20000010
ALL_CHARACTERS_AUTHORIZED = 20000011
ALL_PLAYABLE_AUTHORIZED = ALL_CHARACTERS_AUTHORIZED
ALL_MOVABLE_AUTHORIZED = 20000100
NONE_AUTHORIZED = 20000101

# UI positions
POS_TOP = 30000001
POS_BOTTOM = 30000010
ORIENTATION_HORIZONTAL = 30000011
ORIENTATION_VERTICAL = 30000100


# Running states
RUNNING = 40000001
PAUSED = 40000010
STOPPED = 40000011

# Special constants
NO_PLAYER = 90000001
MODE_RT = 90000002
MODE_TBT = 90000003



import pygamelib.gfx.core as gfx_core
import pygamelib.board_items as board_items
import pygamelib.assets.graphics as graphics
import pygamelib.constants as constants
import pygamelib.base as base

import random
import uuid


class BaseParticle(board_items.Movable):
    """
    Particles are not ready. This is only an early early test.
    *you should not use it*. If you do, don't complain. And if you really want to help,
    interact on Github or Discord.
    Thank you ;)
    """

    def __init__(self, **kwargs):
        board_items.Movable.__init__(self, **kwargs)
        self.size = [1, 1]
        self.name = str(uuid.uuid4())
        self.type = "base_particle"
        self.sprixel = gfx_core.Sprixel(graphics.GeometricShapes.BULLET)
        if "bg_color" in kwargs:
            self.sprixel.bg_color = kwargs["bg_color"]
        if "fg_color" in kwargs:
            self.sprixel.fg_color = kwargs["fg_color"]
        if "model" in kwargs:
            self.sprixel.model = kwargs["model"]
        self.directions = [
            base.Vector2D.from_direction(constants.UP, 1),
            base.Vector2D.from_direction(constants.DLUP, 1),
            base.Vector2D.from_direction(constants.DRUP, 1),
        ]
        self.lifespan = 5
        if "velocity" in kwargs and isinstance(kwargs["velocity"], base.Vector2D):
            self.velocity = kwargs["velocity"]
        else:
            self.velocity = base.Vector2D()
        if "acceleration" in kwargs and isinstance(
            kwargs["acceleration"], base.Vector2D
        ):
            self.acceleration = kwargs["acceleration"]
        else:
            self.acceleration = base.Vector2D()

        for item in ["lifespan", "sprixel", "name", "type", "directions"]:
            if item in kwargs:
                setattr(self, item, kwargs[item])

    def direction(self):
        return random.choice(self.directions)

    def pickable(self):
        """
        A particle is not pickable by default. So that method returns False.
        """
        return False

    def overlappable(self):
        """
        Overlapable always return true. As by definition a particle is overlapable.
        """
        return True



"""
The Game.py module has only one class: Game. It is what could be called the game engine.
It holds a lot of methods that helps taking care of some complex mechanics behind the
curtain.

This module contains the Inventory class.

This module regroup all the specific exceptions of the library.
The idea behind most exceptions is to provide more context and info that the standard
exceptions.

This module contains the Board class.
It is the base class for all levels.

.. autosummary::
   :toctree: .

   Math
   PglException
   PglInvalidLevelException
   PglInvalidTypeException
   PglObjectIsNotMovableException
   PglOutOfBoardBoundException
   Vector2D
   Text
"""
from pygamelib import constants
import math
from colorama import Fore, Back, Style, init

# Initialize terminal colors for colorama.
init()


class Text:
    """
    An object to manipulate and display text in multiple contexts.

    .. versionadded:: 1.2.0

    The Text class is a collection of text formating and display static methods.

    You can either instantiate an object or use the static methods.

    The Text object allow for easy text manipulation through its collection of
    independent attributes. They help to set the text, its style and the foreground and
    background colors.

    The Text object can generate a :class:`~pygamelib.gfx.core.Sprite` to represent
    itself. This is particularly useful to the place text on the game
    :class:`~pygamelib.engine.Board`.

    :param text: The text to manipulate
    :type text: str
    :param fg_color: The foreground color for the text.
    :type fg_color: str
    :param bg_color: The background color for the text.
    :type bg_color: str
    :param style: The style for the text.
    :type style: str
    """

    def __init__(self, text="", fg_color="", bg_color="", style=""):
        self.text = text
        """The text attribute. It needs to be a str."""
        self.fg_color = fg_color
        """The fg_color attribute sets the foreground color. It needs to be a str."""
        self.bg_color = bg_color
        """The bg_color attribute sets the background color. It needs to be a str."""
        self.style = style
        """The style attribute sets the style of the text. It needs to be a str."""
        self.parent = None
        """This object's parent. It needs to be a
        :class:`~pygamelib.board_items.BoardItem`."""
        self._sprite = None
        self._item = None

    def __repr__(self):
        return "".join([self.bg_color, self.fg_color, self.style, self.text, "\x1b[0m"])

    def __str__(self):  # pragma: no cover
        return self.__repr__()

    @staticmethod
    def warn(message):
        """Print a warning message.

        The warning is a regular message prefixed by WARNING in black on a yellow
        background.

        :param message: The message to print.
        :type message: str

        Example::

            base.Text.warn("This is a warning.")
        """
        print(Fore.BLACK + Back.YELLOW + "WARNING" + Style.RESET_ALL + ": " + message)

    @staticmethod
    def fatal(message):
        """Print a fatal message.

        The fatal message is a regular message prefixed by FATAL in white on a red
        background.

        :param message: The message to print.
        :type message: str

        Example::

            base.Text.fatal("|x_x|")
        """
        print(
            Fore.WHITE
            + Back.RED
            + Style.BRIGHT
            + "FATAL"
            + Style.RESET_ALL
            + ": "
            + message
        )

    @staticmethod
    def info(message):
        """Print an informative message.

        The info is a regular message prefixed by INFO in white on a blue background.

        :param message: The message to print.
        :type message: str

        Example::

            base.Text.info("This is a very informative message.")
        """
        print(Fore.WHITE + Back.BLUE + "INFO" + Style.RESET_ALL + ": " + message)

    @staticmethod
    def debug(message):
        """Print a debug message.

        The debug message is a regular message prefixed by INFO in blue on a green
        background.

        :param message: The message to print.
        :type message: str

        Example::

            base.Text.debug("This is probably going to success, eventually...")
        """
        print(
            Fore.BLUE
            + Back.GREEN
            + Style.BRIGHT
            + "DEBUG"
            + Style.RESET_ALL
            + ": "
            + message
        )

    @staticmethod
    def print_white_on_red(message):
        """Print a white message over a red background.

        :param message: The message to print.
        :type message: str

        Example::

            base.Text.print_white_on_red("This is bright!")
        """
        print(Fore.WHITE + Back.RED + message + Style.RESET_ALL)

    # Colored bright functions
    @staticmethod
    def green_bright(message):
        """
        Return a string formatted to be bright green

        :param message: The message to format.
        :type message: str
        :return: The formatted string
        :rtype: str

        Example::

            print( Text.green_bright("This is a formatted message") )

        """
        return Fore.GREEN + Style.BRIGHT + message + Style.RESET_ALL

    @staticmethod
    def blue_bright(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.BLUE + Style.BRIGHT + message + Style.RESET_ALL

    @staticmethod
    def red_bright(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.RED + Style.BRIGHT + message + Style.RESET_ALL

    @staticmethod
    def yellow_bright(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.YELLOW + Style.BRIGHT + message + Style.RESET_ALL

    @staticmethod
    def magenta_bright(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.MAGENTA + Style.BRIGHT + message + Style.RESET_ALL

    @staticmethod
    def cyan_bright(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.CYAN + Style.BRIGHT + message + Style.RESET_ALL

    @staticmethod
    def white_bright(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.WHITE + Style.BRIGHT + message + Style.RESET_ALL

    @staticmethod
    def black_bright(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.BLACK + Style.BRIGHT + message + Style.RESET_ALL

    # Colored normal functions
    @staticmethod
    def green(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.GREEN + message + Style.RESET_ALL

    @staticmethod
    def blue(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.BLUE + message + Style.RESET_ALL

    @staticmethod
    def red(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.RED + message + Style.RESET_ALL

    @staticmethod
    def yellow(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.YELLOW + message + Style.RESET_ALL

    @staticmethod
    def magenta(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.MAGENTA + message + Style.RESET_ALL

    @staticmethod
    def cyan(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.CYAN + message + Style.RESET_ALL

    @staticmethod
    def white(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.WHITE + message + Style.RESET_ALL

    @staticmethod
    def black(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.BLACK + message + Style.RESET_ALL

    # Colored dim function
    @staticmethod
    def green_dim(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.GREEN + Style.DIM + message + Style.RESET_ALL

    @staticmethod
    def blue_dim(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.BLUE + Style.DIM + message + Style.RESET_ALL

    @staticmethod
    def red_dim(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.RED + Style.DIM + message + Style.RESET_ALL

    @staticmethod
    def yellow_dim(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.YELLOW + Style.DIM + message + Style.RESET_ALL

    @staticmethod
    def magenta_dim(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.MAGENTA + Style.DIM + message + Style.RESET_ALL

    @staticmethod
    def cyan_dim(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.CYAN + Style.DIM + message + Style.RESET_ALL

    @staticmethod
    def white_dim(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.WHITE + Style.DIM + message + Style.RESET_ALL

    @staticmethod
    def black_dim(message):
        """
        This method works exactly the way green_bright() work with different color.
        """
        return Fore.BLACK + Style.DIM + message + Style.RESET_ALL


class PglInvalidTypeException(Exception):
    """
    Exception raised for invalid types.
    """

    def __init__(self, message):
        self.message = message


class HacInvalidTypeException(PglInvalidTypeException):
    """A simple forward to PglInvalidTypeException
    """


class PglException(Exception):
    """
    Exception raised for non specific errors in the pygamelib.
    """

    def __init__(self, error, message):
        self.error = error
        self.message = message


class HacException(PglException):
    """A simple forward to PglException
    """


class PglOutOfBoardBoundException(Exception):
    """
    Exception for out of the board's boundaries operations.
    """

    def __init__(self, message):
        self.message = message


class PglOutOfItemBoundException(Exception):
    """
    Exception for out of the item's boundaries operations.
    """

    def __init__(self, message):
        self.message = message


class HacOutOfBoardBoundException(PglOutOfBoardBoundException):
    """Simple forward to PglOutOfBoardBoundException
    """


class PglObjectIsNotMovableException(Exception):
    """
    Exception raised if the object that is being moved is not a subclass of Movable.
    """

    def __init__(self, message):
        self.message = message


class HacObjectIsNotMovableException(PglObjectIsNotMovableException):
    """Simple forward to PglObjectIsNotMovableException
    """


class PglInvalidLevelException(Exception):
    """
    Exception raised if a level is not associated to a board in Game().
    """

    def __init__(self, message):
        self.message = message


class HacInvalidLevelException(PglInvalidLevelException):
    """Forward to PglInvalidLevelException
    """


class PglInventoryException(Exception):
    """
    Exception raised for issue related to the inventory.
    The error is an explicit string, and the message explains the error.
    """

    def __init__(self, error, message):
        self.error = error
        self.message = message


class HacInventoryException(PglInventoryException):
    """Forward to PglInventoryException.
    """


class Vector2D:
    """A 2D vector class.

    .. versionadded:: 1.2.0

    Contrary to the rest of the library Vector2D uses floating point numbers for its
    coordinates/direction/orientation. However since the rest of the library uses
    integers, the numbers are rounded to 2 decimals.
    You can alter that behavior by increasing or decreasing (if you want integer for
    example).

    Vector2D use the row/column internal naming convention as it is easier to visualize
    For learning developers. If it is a concept that you already understand and are
    more familiar with the x/y coordinate system you can also use x and y.

     - x is equivalent to column
     - y is equivalent to row

    Everything else is the same.

    Vectors can be printed and supports basic operations:

     - addition
     - substraction
     - multiplication

    Let's elaborate a bit more on the multiplication. The product behaves in 2 different
    ways:

    If you multiply a vector with a scalar (int or float), the return value is a
    Vector2D with each vector component multiplied by said scalar.

    If you multiply a Vector2D with another Vector2D you ask for the the cross
    product of vectors. This is an undefined mathematical operation in 2D as the
    cross product is supposed to be perpendicular to the 2 other vectors (along the
    z axis in our case). Since we don't have depth (z) in 2D, this will return the
    magnitude of the signed cross product of the 2 vectors.

    Example of products::

        v1 = base.Vector2D(1,2)
        v2 = base.Vector2D(3,4)
        # This returns -2
        mag = v1 * v2
        # This returns a Vector2D with values (-1, -2)
        inv = v1 * -1
        # This return a Vector2D with values (2.85, 3.8) or 95% of v2
        dim = v2 * 0.95

    :param row: The row/y parameter.
    :type row: int
    :param column: The column/x parameter.
    :type column: int

    Example::

        gravity = Vector2D(9.81, 0)
        # Remember that minus on row is up.
        speed = Vector2D(-0.123, 0.456)
        # In that case you might want to increase the rounding precision
        speed.rounding_precision = 3
    """

    def __init__(self, row=0.0, column=0.0):
        # column is x and row is y
        self.__row = row
        self.__column = column
        self.rounding_precision = 2
        """The rounding_precision attributes is used when vectors values are calculated
        and the result rounded for convenience. It can be changed anytime to increase or
        decrease the precision anytime."""

    def __repr__(self):
        return f"{self.__class__.__name__} ({self.__row}, {self.__column})"

    def __str__(self):  # pragma: no cover
        return self.__repr__()

    def __add__(self, other):
        row = round(self.__row + other.row, self.rounding_precision)
        column = round(self.__column + other.column, self.rounding_precision)
        return Vector2D(row, column)

    def __sub__(self, other):
        row = round(self.__row - other.row, self.rounding_precision)
        column = round(self.__column - other.column, self.rounding_precision)
        return Vector2D(row, column)

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return Vector2D(
                round(self.__row * other, self.rounding_precision),
                round(self.__column * other, self.rounding_precision),
            )
        if isinstance(other, Vector2D):
            return round(
                self.row * other.column - self.column * other.row,
                self.rounding_precision,
            )

    def __eq__(self, other):
        if other.row == self.__row and other.column == self.__column:
            return True
        return False

    @property
    def row(self):
        """
        The row component of the vector.
        """
        return self.__row

    @row.setter
    def row(self, value):
        if isinstance(value, (int, float)):
            self.__row = round(value, self.rounding_precision)
        else:
            raise PglInvalidTypeException("Vector2D.row needs to be an int or a float.")

    @property
    def y(self):
        """
        y is an alias for row.
        """
        return self.row

    @y.setter
    def y(self, value):
        self.row = value

    @property
    def column(self):
        """
        The column component of the vector.
        """
        return self.__column

    @column.setter
    def column(self, value):
        if isinstance(value, (int, float)):
            self.__column = round(value, self.rounding_precision)
        else:
            raise PglInvalidTypeException(
                "Vector2D.column needs to be an int or a float."
            )

    @property
    def x(self):
        """
        x is an alias for column.
        """
        return self.column

    @x.setter
    def x(self, value):
        self.column = value

    def length(self):
        """
        Returns the length of a vector.

        :rtype: float

        Example::

            if speed.length() == 0.0:
                print('We are not moving... at all...')
        """
        return round(
            math.sqrt(self.row ** 2 + self.column ** 2), self.rounding_precision
        )

    def unit(self):
        """Returns a normalized unit vector.

        :returns: A unit vector
        :rtype: :class:`~pygamelib.base.Vector2D`

        Example::

            gravity = Vector2D(9.81, 0)
            next_position = item.position_as_vector() + gravity.unit()
        """
        if self.length() == 0.0:
            return Vector2D()
        return Vector2D(
            round(self.__row / self.length(), self.rounding_precision),
            round(self.__column / self.length(), self.rounding_precision),
        )

    @classmethod
    def from_direction(cls, direction, step):
        """Build and return a Vector2D from a direction.

        Directions are from the constants module.

        :param direction: A direction from the constants module.
        :type direction: int
        :param step: The number of cell to cross in one movement.
        :type step: int

        Example::

            v2d_up = Vector2D.from_direction(constants.UP, 1)
        """
        if direction == constants.NO_DIR:
            return cls(0, 0)
        elif direction == constants.UP:
            return cls(-step, 0)
        elif direction == constants.DOWN:
            return cls(+step, 0)
        elif direction == constants.LEFT:
            return cls(0, -step)
        elif direction == constants.RIGHT:
            return cls(0, +step)
        elif direction == constants.DRUP:
            return cls(-step, +step)
        elif direction == constants.DRDOWN:
            return cls(+step, +step)
        elif direction == constants.DLUP:
            return cls(-step, -step)
        elif direction == constants.DLDOWN:
            return cls(+step, -step)


class Math(object):
    """The math class regroup math functions required for game development.

    .. versionadded:: 1.2.0

    For the moment there is only static methods in that class but it will evolve in the
    future.

    """

    def __init__(self):
        super().__init__()

    @staticmethod
    def intersect(row1, column1, width1, height1, row2, column2, width2, height2):
        """This function check if 2 rectangles intersect.

        The 2 rectangles are defined by their positions (row, column) and dimension
        (width and height).

        :param row1: The row of the first rectangle
        :type row1: int
        :param column1: The column of the first rectangle
        :type column1: int
        :param width1: The width of the first rectangle
        :type width1: int
        :param height1: The height of the first rectangle
        :type height1: int
        :param row2: The row of the second rectangle
        :type row2: int
        :param column2: The column of the second rectangle
        :type row2: int
        :param width2: The width of the second rectangle
        :type width2: int
        :param height2: The height of the second rectangle
        :type height2: int
        :returns: A boolean, True if the rectangles intersect False, otherwise.

        Example::

            if intersect(projectile.row, projectile.column, projectile.width,
                         projectile.height, bady.row, bady.column, bady.width,
                         bady.height):
                projectile.hit([bady])
        """
        # Shortcut: if they are at the same position they obviously intersect
        if row1 == row2 and column1 == column2:
            return True
        return (
            max(row1, row1 + height1 - 1) >= min(row2, row2 + height2 - 1)
            and min(row1, row1 + height1 - 1) <= max(row2, row2 + height2 - 1)
            and max(column1, column1 + width1 - 1) >= min(column2, column2 + width2 - 1)
            and min(column1, column1 + width1 - 1) <= max(column2, column2 + width2 - 1)
        )

    @staticmethod
    def distance(row1, column1, row2, column2):
        """Return the euclidian distance between to points.

        Points are identified by their row and column.
        If you want the distance in number of cells, you need to round the result (see
        example).

        :param row1: the row number (coordinate) of the first point.
        :type row1: int
        :param column1: the column number (coordinate) of the first point.
        :type column1: int
        :param row2: the row number (coordinate) of the second point.
        :type row2: int
        :param column2: the column number (coordinate) of the second point.
        :type column2: int
        :return: The distance between the 2 points.
        :rtype: float

        Example::

            distance = round(base.Math.distance(player.row,
                                            player.column,
                                            npc.row,
                                            npc.column)
        """
        return math.sqrt((column2 - column1) ** 2 + (row2 - row1) ** 2)



import setuptools
import pygamelib.constants as constants

try:
    from pipenv.project import Project
    from pipenv.utils import convert_deps_to_pip
except ImportError:
    print("Please install pipenv first. See: https://github.com/pypa/pipenv")

with open("README.md", "r") as fh:
    long_description = fh.read()

# Compatibility layer between Pipenv and Pip requirements.txt
# See https://github.com/pypa/pipenv/issues/209
pipfile = Project(chdir=False).parsed_pipfile
requirements_path = convert_deps_to_pip(pipfile["packages"])

INSTALL_PACKAGES = open(requirements_path).read().splitlines()

setuptools.setup(
    name="pygamelib",
    version=constants.PYGAMELIB_VERSION,
    author="Arnaud Dupuis",
    author_email="8bitscoding@gmail.com",
    description="A small game development framework for teaching programming.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=INSTALL_PACKAGES,
    url="https://www.pygamelib.org",
    packages=setuptools.find_packages(),
    scripts=["pgl-editor.py", "pgl-board-tester.py"],
    keywords=["game", "development", "beginner", "console", "terminal"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    project_urls={
        "Documentation": "https://pygamelib.readthedocs.io/en/latest/index.html",  # noqa: E501
        "Guides & Tutorials": "https://astro.hyrul.es/guides/pygamelib/",
        "Source": "https://github.com/arnauddupuis/pygamelib",
        "Tracker": "https://github.com/arnauddupuis/pygamelib/issues",
        "Wiki": "https://github.com/arnauddupuis/pygamelib/wiki",
        "Tech blog": "https://8bitscoding.io/",
        "Release Notes": "https://8bitscoding.io/2020/08/30/pygamelib-v1-2-0-release-notes/",  # noqa: E501
    },
    python_requires=">=3.6",
    obsolotes="hac-game-lib",
)
